"""
 Module to read AMBER residue lib
"""

import re
import sys

class ResidueLib():

    def __init__(self, library_path):
        try:
            fh = open(library_path, "r")

        except IOError:
            print ("ERROR: unable to open residue library" + library_path, file=sys.stderr)
            sys.exit(1)

        line = fh.readline()
        line = fh.readline()

        self.residues = {}
        at_group = False
        ch_group = False
        im_group = False
        res = ResidueDef()
        for line in fh:
            line = line.replace("\n", "").replace("\r", "")
            if line == '':
                continue

            elif line == "DONE":
                self.residues[res.id] = res
                at_group = False
                ch_group = False
                im_group = False
                res = ResidueDef()

            elif line == 'CHARGE':
                at_group = False
                ch_group = True

            elif line == 'IMPROPER':
                ch_group = False
                im_group = True

            elif re.match(' (...)  INT', line):
                resid_str = re.match(' (...)  INT', line)
                res.id = resid_str.group(1)
                at_group = True

            elif at_group:
                data = line.split()
                if len(data) < 11:
                    continue
                at = AtomDef(data)
                res.ats.append(at)

            elif ch_group:
                for c in line.split():
                    res.charges.append(c)

            elif im_group:
                res.improper.append(line)

class ResidueDef():

    def __init__(self):
        self.id = ''
        self.name = ''
        self.ats = ['']
        self.improper = []
        self.charges = []

class AtomDef():

    def __init__(self, data):
        self.id = data[1]
        self.type = data[2]
        self.branch = data[3]
        self.link = [int(data[4]), int(data[5]), int(data[6])]
        self.geom = [float(data[7]), float(data[8]), float(data[9])]
        self.coord = [[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]]
        self.ch = float(data[10])

    def __str__(self):
        return self.id



"""Module to manage structure, based on BioPython Bio.PDB
"""
import sys
import warnings
from Bio import BiopythonWarning
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
from Bio.PDB.MMCIFParser import MMCIFParser
from Bio.PDB.PDBIO import PDBIO
from Bio.PDB.PDBList import PDBList
from Bio.PDB.PDBParser import PDBParser
from Bio.PDB.Atom import Atom
from Bio.PDB.parse_pdb_header import parse_pdb_header

import biobb_model.structure_manager.model_utils as mu

class StructureManager():
    """Main Class wrapping Bio.PDB structure object
    """

    def __init__(self, input_pdb_path):
        """Class constructor. Sets an empty object and loads a structure
        according to parameters

        Args:
            **input_pdb_path** (str): path to input structure either in pdb or
            mmCIF format. Format is taken from file extension.
            Alternatively **pdb:pdbId** fetches the mmCif file from RCSB

        Object structure:
            {
                "input_format" (str): Format of input file pdb|cif
                "model_type" (int): Guessed model type when num. models > 1 NMR|Bunit
                "num_ats" (int): Total Number of atoms
                "nmodels" (int): Number of Models
                "chain_ids" (List): Chain composition as
                    [chain_id:{}]
                "modified" (Boolean): Flag to indicated that structure has been modified
                "all_residues" (list): List of pointer to Bio.PDB.Residue objects, ordered
                    acording to input file
                "num_res" (int): Number of residues
                "res_insc" (int): Number of residues with insertion codes
                "res_hetats" (int): Number of residues flagged as HETATM
                "res_ligands" (int): Number of non water residues flagged as HETATM
                "num_wat" (int): Number of water residues
                "ca_only" (boolead): Flag to indicate a possible CA-only structure
                "bck_breaks_list" []: List of backbone breaks as [r1,r2] tuples
                "wrong_link_list" []: List of wrongly connected residues as [r1, r2, r3] tuples
                    where r1-r2 is the actual link whereas r1-r3 was expected
                "not_link_seq_list" []: List of sequence consecutive residues that are not linked as
                    [r1,r2] tuples. Most probable due to bond distance above the Covalent distance
                "modified_residue_list" []: List of residues being connected HETAM as PDB.Bio.Residue
                "backbone_links" []: List of found canonical backbone links as [at1, at2] tuples, according to
                    a distance criterium
                "residue_link_to" {}: Dict r1:r2 for pairs of linked residues in the N-term to C-term direction
                "cis_backbone_list" []: List of cis backbone links as [r1,r2,dihedral] tuples
                "lowtrans_backbone_list" []: List of trans backbone links with reduced dihedral as [r1,r2,dihedral] tuples

        """

        self.input_format = ''
        self.model_type = ''
        self.num_ats = 0
        self.nmodels = 0
        self.chain_ids = []
        self.modified = False
        self.all_residues = []

# TODO support biounits
        if "pdb:"in input_pdb_path:
            pdbl = PDBList(pdb='tmpPDB')

            try:
                input_pdb_path = input_pdb_path[4:].upper()
                real_pdb_path = pdbl.retrieve_pdb_file(input_pdb_path, file_format='mmCif')
                parser = MMCIFParser()
                self.input_format = 'cif'

            except IOError:
                sys.stderr.write("ERROR: fetching structure at {}\n".format(input_pdb_path))
#                print ('ERROR: fetching structure at {}'.format(input_pdb_path), file=sys.stderr)
                sys.exit(2)
        else:
            real_pdb_path = input_pdb_path
            if real_pdb_path is tuple:
                real_pdb_path = real_pdb_path[0]
        print(real_pdb_path)
        if '.pdb' in real_pdb_path:
            parser = PDBParser(PERMISSIVE=1)
            self.input_format = 'pdb'
        elif '.cif' in real_pdb_path:
            parser = MMCIFParser()
            self.input_format = 'cif'
        else:
            print ('ERROR: unknown filetype', file=sys.stderr)
            sys.exit(2)
        try:
            warnings.simplefilter('ignore', BiopythonWarning)
            self.st = parser.get_structure('st', real_pdb_path)
            if self.input_format == 'pdb':
                self.headers = parse_pdb_header(real_pdb_path)
            else:
                self.headers = MMCIF2Dict(real_pdb_path)

        except OSError:
            print ("#ERROR: parsing PDB", file=sys.stderr)
            sys.exit(2)
        # Add .index field for correlative, unique numbering of residues
        self.residue_renumbering()

    #Atom renumbering for mmCIF, PDB uses atom number in file
        if self.input_format == 'cif':
            self.atom_renumbering()

    #checking models type according to RMS among models
        self.nmodels = len(self.st)
        if self.nmodels > 1:
            self.models_type = mu.guess_models_type(self.st)
        else:
            self.models_type = 0
    #Identify chains and guess type (protein, dna, ...)
        self.set_chain_ids()
    #Calc general stats (num residues, atoms, etc)
        self.calc_stats()

    def residue_renumbering(self):
        """Sets the Bio.PDB.Residue.index attribute to residues for a unique,
        consecutive, residue number, and generates the corresponding list
        in **all_residues**
        """
        i = 1
        self.all_residues=[]
        for r in self.st.get_residues():
            r.index = i
            self.all_residues.append(r)
            i += 1

    def atom_renumbering(self):
        """Sets  Bio.PDB.Atom.serial_number for all atoms in the structure, overriding original if any
        """
        i = 1
        for at in self.st.get_atoms():
            at.serial_number = i
            if hasattr(at, 'selected_child'):
                at.selected_child.serial_number = i
            i += 1

    def calc_stats(self):
        """Calculates general statistics about the structure, and guesses whether
        it is a CA-only structure
        """
        nr = 0
        na = 0
        hi =     0
        wi = 0
        insi = 0
        for r in self.st.get_residues():
            nr += 1
            if mu.is_wat(r):
                wi += 1
            if mu.is_hetatm(r):
                hi += 1
            if mu.has_ins_code(r):
                insi += 1
            na += len(r.get_list())
        self.num_res = nr
        self.num_ats = na
        self.res_insc = insi
        self.res_hetats = hi
        self.res_ligands = hi-wi
        self.num_wat = wi
        # Detecting whether it is a CA-only structure
        # num_ats should be much larger than num_res
        # waters removed
        # Taking polyGly as a lower limit
        self.ca_only = na - wi < (nr-wi)*4

    def get_ins_codes(self):
        """Makes a list with residues having insertion codes"""

        self.ins_codes_list=[]
        for r in self.st.get_residues():
            if mu.has_ins_code(r):
                self.ins_codes_list.append(r)

    def check_missing_atoms(self, valid_codes, residue_data):
        """ Makes a **list of missing atoms** in the structure

            Args:
                **valid_codes**: Valid residue 3-letter codes
                **residue_data**: Atom composition per residue taken
                    from residue library
            Returns:
                List of residues with missing atoms, as a tuples
                ["r",{"backbone":[atoms_list],"side":[atoms_list]}]
        """

        miss_at_list = []
        for r in self.st.get_residues():
            if r.get_resname() in valid_codes and not mu.is_hetatm(r):
                miss_at = mu.check_all_at_in_r(r, residue_data[r.get_resname().replace(' ','')])
                if len(miss_at) > 0:
                    miss_at_list.append([r,miss_at])
        return miss_at_list

    def get_missing_side_chain_atoms(self, valid_codes, residue_data):
        """
            returns list of residues with complete backbone but with missing
            side chain atoms

            Args:
                **valid_codes**: list of valid 3-letter aminoacid codes
                **residue_data**: atom composition per residue taken
                    from residue library
            Returns:
                List of residues with missing atoms, as a tuples like
                ["r",[atoms_list]]
        """
        miss_side = []
        for res in self.check_missing_atoms(valid_codes, residue_data):
            [r,at_list]=res
            if not len(at_list['backbone']):
                miss_side.append([r,at_list['side']])
        return miss_side

    def get_missing_backbone_atoms(self, valid_codes, residue_data):
        """
            returns list of residues with missing backbone atoms

            Args:
                **valid_codes**: list of valid 3-letter aminoacid codes
                **residue_data**: atom composition per residue taken
                    from residue library
            Returns:
                List of residues with missing atoms, as tuples
                ["r",[atoms_list]]
        """
        miss_bck = []
        for res in self.check_missing_atoms(valid_codes, residue_data):
            [r,at_list]=res
            if at_list['backbone']:
                miss_bck.append([r,at_list['backbone']])
        return miss_bck

    def get_bck_breaks(self):
        """
            Determines several backbone anomalies
            1. Modified residues
            2. Backbone breaks
            3. Backbone links not corresponding to sequence
            4. Too long residue links
        """
        self.bck_breaks_list=[]
        self.wrong_link_list=[]
        self.not_link_seq_list=[]
        self.modified_residue_list=[]
        for i in range(0,len(self.all_residues)-1):
            r1 = self.all_residues[i]
            r2 = self.all_residues[i+1]
            if not mu.same_chain(r1,r2):
                continue
            if mu.is_hetatm(r1) or mu.is_hetatm(r2):
                if r1 in self.residue_link_to:
                    if self.residue_link_to[r1] == r2:
                        if mu.is_hetatm(r1):
                            self.modified_residue_list.append(r1)
                else:
                    continue
            #Skip NA #TODO include NA Backbone
            if self.chain_ids[r1.get_parent().id] != mu.PROTEIN:
                continue

            if r1 not in self.residue_link_to:
                self.bck_breaks_list.append([r1,r2])
                if mu.seq_consecutive(r1,r2):
                    d=0.
                    if 'N' in r1 and 'C' in r2:
                        d = r1['N']-r2['C']
                    self.not_link_seq_list.append([r1,r2,d])


            else:
                if r2 != self.residue_link_to[r1]:
                    self.wrong_link_list.append([r1,self.residue_link_to[r1],r2])

    def check_backbone_connect(self, backbone_atoms, COVLNK):
        """
        Determines backbone links usign a distance criterium and produces a dict with
        link relationships in the N-term to C-term direction

        Args:
            beckbone_atoms: atoms to be considered as backbone
            COVLNK: Threshold distance for a covalent bond
        """
        self.backbone_links = mu.get_backbone_links(self.get_structure(), backbone_atoms, COVLNK)
        self.residue_link_to = {}
        for lnk in self.backbone_links:
            [at1, at2] = lnk
            self.residue_link_to[at1.get_parent()]=at2.get_parent()
            # diff chain
            # no n->n+1

    def check_cis_backbone(self, backbone_atoms, COVLNK):
        """
        Determines omega dihedrals for two bound residues and classifies them
        as normal trans, low angle trans, and cis

        Args:
            backbone_atoms: atoms to be considered as backbone
            COVLNK: Distance thresjold for a covalent bond

        Internal parameters:
            CISTHRES (20): Max dihedral for cis bonds
            TRANSTHRES (160): Min dihedral for trans bonds
        """
        CISTHRES = 20  #TODO check vaules withpdb checking
        TRANSTHRES = 160
        if not hasattr(self,'backbone_links'):
            self.backbone_links = mu.get_backbone_links(self.get_structure(), backbone_atoms, COVLNK)
        self.cis_backbone_list = []
        self.lowtrans_backbone_list = []
        for lnk in self.backbone_links:
            [at1, at2] = lnk
            r1 = at1.get_parent()
            r2 = at2.get_parent()
            if 'CA' in r1 and 'CA' in r2:
                dih =  mu.calc_bond_dihedral(r1['CA'],r1['C'],r2['N'],r2['CA'])
                if abs(dih) < CISTHRES:
                    self.cis_backbone_list.append([r1,r2,dih])
                elif abs(dih) < TRANSTHRES:
                    self.lowtrans_backbone_list.append([r1,r2,dih])

    def get_stats(self):
        """
         Returns a dict with calculates statistics

         Returns:
            Dict as {}
        """
        return {
            'nmodels': self.nmodels,
            'nchains': len(self.chain_ids),
            'chain_ids': self.chain_ids,
            'num_res': self.num_res,
            'num_ats': self.num_ats,
            'res_insc': self.res_insc,
            'res_hetats': self.res_hetats,
            'res_ligands': self.res_ligands,
            'num_wat': self.num_wat,
            'ca_only': self.ca_only
        }
    def print_headers(self):
        """
        Prints selected components from structure headers
        """
        if self.input_format== 'cif':
            print (' PDB id: {}'.format(self.headers['_entry.id']))
            print (' Title: {}'.format(self.headers['_struct.title']))
            print (' Experimental method: {}'.format(self.headers['_exptl.method']))
            print (' Keywords: {}'.format(self.headers['_struct_keywords.pdbx_keywords']))
            if '_refine_hist.d_res_high' in self.headers:
                print (' Resolution: {} A'.format(self.headers['_refine_hist.d_res_high']))
        else:
            print ('Title: {}'.format(self.headers['name']))
            print ('Experimental method: {}'.format(self.headers['structure_method']))
            if 'keywords' in self.headers:
                print ('Keywords: {}'.format(self.headers['keywords']))
            if 'resolution' in self.headers:
                print ('Resolution: {} A'.format(self.headers['resolution']))

    def print_stats(self, prefix=''):
        """
        Prints statistics to stdout

        Args:
            prefix: Text prefix to prepend to printed data
        """
        stats = self.get_stats()

        print ('{} Num. models: {}'.format(prefix, stats['nmodels']))
        chids=[]
        for ch_id in sorted(stats['chain_ids']):
            if isinstance(stats['chain_ids'][ch_id],list):
                chids.append('{}: Unknown '.format(ch_id))
            else:
                chids.append('{}: {}'.format(ch_id, mu.chain_type_labels[stats['chain_ids'][ch_id]]))
        print ('{} Num. chains: {} ({})'.format(prefix, stats['nchains'], ', '.join(chids)))
        print ('{} Num. residues:  {}'.format(prefix, stats['num_res']))
        print ('{} Num. residues with ins. codes:  {}'.format(prefix, stats['res_insc']))
        print ('{} Num. HETATM residues:  {}'.format(prefix, stats['res_hetats']))
        print ('{} Num. ligands or modified residues:  {}'.format(prefix, stats['res_ligands']))
        print ('{} Num. water mol.:  {}'.format(prefix, stats['num_wat']))
        print ('{} Num. atoms:  {}'.format(prefix, stats['num_ats']))
        if stats['ca_only']:
            print ('Possible CA-Only structure')

    def get_structure(self):
        """
        Method pointing to the enclosed Bio.PDB.Structure object

        Returns:
            a Bio.PDB.Structure object
        """
        return self.st

    def save_structure(self, output_pdb_path):
        """
        Saves structure on disk in PDB format

        Args:
            output_pdb_path: OS path to the output file

        Errors:
            OSError: Error saving the file
        """
        if output_pdb_path:
            pdbio = PDBIO()
            pdbio.set_structure(self.st)
            try:
                pdbio.save(output_pdb_path)

            except OSError:
                sys.stderr.write ("#ERROR: unable to save PDB data on " + output_path)
        else:
            sys.stderr.write ("Error: output_pdb_path not provided \n")
            sys.exit(1)

# Methods to modify structure
    def select_model(self, nm):
        """
        Selects one model and delete the others from the structure. Model is
        renumbered to model 1

        Args:
            nm: Model number to keep
        """
        ids = []
        for md in self.st.get_models():
            ids.append(md.id)
        for i,v in enumerate(ids):
            if i != nm-1:
                self.st.detach_child(v)
        if nm != 1:
            self.st[nm-1].id = 0
        self.nmodels = 1
        self.models_type = 0
        self.residue_renumbering()
        self.atom_renumbering()
        self.set_chain_ids()
        self.modified=True

    def has_models(self):
        """
        Shotcut method to check whether the structure has more than one model

        Returns: Boolean
        """
        return self.nmodels>1

    def set_chain_ids(self):
        """
        Identifies and sets the chain ids, guessing its nature (protein, dna, rna, ...)
        """
        self.chain_ids = {}
        for ch in self.st[0].get_chains():
            self.chain_ids[ch.id] = mu.guess_chain_type(ch)
        self.chains_ids = sorted(self.chain_ids)

    def select_chains(self, select_chains):
        """
        Select one or more chains and remove the remaining.
        Args:
            select_chains: Comma separated chain ids
        """
        if not self.chain_ids:
            self.set_chain_ids()
        ch_ok = select_chains.split(',')
        for ch in ch_ok:
            if not ch in self.chain_ids:
                sys.stderr.write ('Error: requested chain {} not present'.format(ch))
                select_chains = ''
        for md in self.st:
            for ch in self.chain_ids:
                if ch not in ch_ok:
                    self.st[md.id].detach_child(ch)
        self.set_chain_ids()
        self.atom_renumbering()
        self.residue_renumbering()
        self.modified=True

    def select_altloc_residue(self, res, to_fix):
        """
        Selects one alternative conformation when altloc exists. Selection is
        done on the occupancy basis or from the conformation id.
        All relevant atoms in the residue **res** at modified in the same way.
        Triggers **modified** flag.

        Args:
            res: residue (as Bio.PDB.Residue)
            to_fix: atoms in residue to fix as
                {'ats':[atom_list],'select':'occupancy|conf_id'
        """
        for at in to_fix['ats']:
            if to_fix['select'].lower() == 'occupancy':
                newat = at.selected_child
            else:
                if to_fix['select'] in at.child_dict.keys():
                    newat = at.child_dict[to_fix['select']]
                else:
                    print ('Warning: unknown alternative {} in {}'.format(to_fix['select'],mu.atom_id(at)), file=sys.stderr)
                    continue
            newat.disordered_flag = 0
            newat.altloc = ' '
            res.detach_child(at.id)
            res.add (newat)
        res.disordered = 0
        self.modified=True

    def remove_residue(self, r):
        """
        Removes residue **r** from the structure. Triggers **modified** flag
        and atom and residue renumbering
        """
        mu.remove_residue(r)
        self.atom_renumbering()
        self.residue_renumbering()
        self.modified=True

    def fix_side_chain(self,r_at, res_library):
        """
        Fix missing side chain atoms in given residue. Triggers **modified** flag

        Args:
            **r_at**: tuple as [Bio.PDB.Residue, [list of atom ids]]
            **res_library**: Residue Library as structure_manager.residue_lib_manager.ResidueLib
        """
        for at_id in r_at[1]:
            print ("  Adding new atom " + at_id)
            if at_id == 'CB':
                coords = mu.buildCoordsCB(r_at[0])
            else:
                coords = mu.buildCoordsOther(r_at[0], res_library, r_at[0].get_resname(), at_id)

            r_at[0].add(Atom(
                at_id,
                coords,
                99.0,
                1.0,
                ' ',
                ' ' + at_id + ' ',
                0,
                at_id[0:1]
                ))

        self.modified=True



"""
  Module to manage command line for simple parameter input
"""
import argparse

def _get_input (value, prompt_text):
    while value is None or value == '':
        value = input (prompt_text)
    if isinstance(value, str):
        value = value.replace(' ', '')
    return value

class ParamInput():
    def __init__(self, prefix, opt_value, non_interactive=True):
        self.options=[]
        self.opt_value=opt_value
        self.non_interactive = non_interactive
        self.prefix = prefix

    def add_option(self, label, opt_list, case=False, opt_type='list', multiple=False, min_val=0, max_val=0, list2=None):
        self.options.append({
            'label':label,
            'opt_list':opt_list,
            'case':case,
            'type':opt_type,
            'min':min_val,
            'max':max_val,
            'multiple':multiple,
            'list2':list2
        })

    def add_option_all(self):
        self.add_option('all', ['All'])

    def add_option_none(self):
        self.add_option('none', ['None'])

    def add_option_yes_no(self):
        self.add_option('yes', ['Yes'])
        self.add_option('nos', ['No'])

    def run(self):
        prompt_str = self.prefix
        if len(self.options):
            opt_strs=[]
            for opt in self.options:
                if opt['type'] == 'list':
                    opt_strs.append(','.join(opt['opt_list']))
                elif opt['type'] == 'int' or opt['type'] == 'float':
                    if opt['min'] != 0 or opt['max'] != 0:
                        opt_strs.append('{} - {}'.format(opt['min'], opt['max']))
                elif opt['type'] == 'input':
                    opt_strs.append('')
                elif opt['type'] == 'pair_list':
                    str_opt = []
                    for op in opt['opt_list']:
                        str_opt.append('{}:[{}]'.format(op,'|'.join(opt['list2'])))
                    opt_strs.append(','.join(str_opt))
                else:
                    opt_strs.append('?')
            prompt_str += ' ('+' | '.join(opt_strs)+')'
        prompt_str += ': '
# No options, simple input
        if not len(self.options):
            if not self.non_interactive:
                self.opt_value = _get_input(self.opt_value, prompt_str)
            return self.opt_value
# Check alternative options
        input_ok = False
        while not input_ok:
            if not self.non_interactive:
                self.opt_value = _get_input(self.opt_value, prompt_str)
            iopt=0
            if self.opt_value is None:
                return ['error','']
            else:
                while not input_ok and iopt < len(self.options):
                    opt = self.options[iopt]
                    if (opt['type'] == 'list' or opt['type'] == 'pair_list') and isinstance(self.opt_value, str):
                        if not opt['multiple']:
                            values = [self.opt_value]
                        else:
                            values = self.opt_value.split(',')
                        group_ok = True
                        for val in values:
                            if opt['type']=='pair_list':
                                val_sp= val.split(':')
                                val = val_sp[0]
                            if opt['case'] == 'upper':
                                val=val.upper()
                                group_ok = group_ok and val in opt['opt_list']
                            elif opt['case'] == 'lower':
                                val=val.lower()
                                group_ok = group_ok and val in opt['opt_list']
                            elif opt['case'] == 'sensitive':
                                group_ok = group_ok and val in opt['opt_list']
                            else:
                                group_ok = group_ok and val.lower() in list(map(lambda x: x.lower(),opt['opt_list']))
                            if opt['type']=='pair_list':
                                group_ok = group_ok and val_sp[1] in opt['list2']
                        input_ok = group_ok
                    elif opt['type'] == 'int':
                        self.opt_value = int(self.opt_value)
                        input_ok = self.opt_value >= opt['min'] and self.opt_value <= opt['max']
                    if not input_ok:
                        iopt +=1
                if not input_ok:
                    print ("Input not valid")
                    if self.non_interactive:
                        self.options.append({'label':'error'})
                        input_ok=True
                    else:
                        self.opt_value = ''

        return [self.options[iopt]['label'],self.opt_value]


class Dialog():
    def __init__(self):
        self.options={}

    def add_option (self, command, opt_prompt, opt_dest, opt_help, opt_type = str):
        self.options[command] = {
            'command':command,
            'prompt': opt_prompt,
            'dest':   opt_dest,
            'help':   opt_help,
            'type':   opt_type
        }

    def get_parameter (self, command, opts):
        dialog = self.options[command]
        if not dialog:
            print ("Error: no dialog available for {}".format(command))
            return 1
        optsParser = argparse.ArgumentParser(prog=dialog['command'])
        optsParser.add_argument(dialog['prompt'], dest=dialog['dest'], help=dialog['help'], type=dialog['type'])
        return vars(optsParser.parse_args(opts))

    def get_dialog(self,command):
        return self.options[command]

    def exists(self,command):
        return command in self.options



"""
  Module to manage command line arguments, adapted to the application
"""

import argparse

class CmdLine():
    def __init__(self, defaults):
        self.argparser = argparse.ArgumentParser (
            description='Basic Structure checking based on MDWeb'
        )

        self.argparser.add_argument (
            'command',
            help="Command to execute (help: checkStruc commands) ",
            default="interactive"
        )

        self.argparser.add_argument(
            '-i', '--input',
            dest='input_structure_path',
            help='Input structure. Formats PDB|mmCIF. Remote pdb:{pdbid}'
        )

        self.argparser.add_argument(
            '-o', '--output',
            dest='output_structure_path',
            help='Output structure. Format PDB'
        )

        self.argparser.add_argument(
            '-d', '--debug',
            action="store_true",
            dest="debug",
            help='Print DEBUG information'
        )

        self.argparser.add_argument(
            '--version',
            action='version',
            version="%(prog)s v0.1"
        )

        self.argparser.add_argument(
            '--data_dir',
            dest='data_dir',
            help="Override settings default data dir"
        )

        self.argparser.add_argument(
            '--res_lib',
            dest='res_lib_path',
            help="Override settings default residue library (AMBER prep format)"
        )

        self.argparser.add_argument(
            '--data_lib',
            dest='data_library_path',
            help="Override settings default data library"
        )

        self.argparser.add_argument(
            '--json',
            dest='json_output_path',
            help='Cummulated checking results on a json file'
            )

        self.argparser.add_argument(
            '--quiet',
            action="store_true",
            dest='quiet',
            help='Reduces output, removing labels and progress info'
        )
        self.argparser.add_argument(
            '--check_only',
            action="store_true",
            dest='check_only',
            help='Perform checks, structures is not modified'
        )

        self.argparser.add_argument(
            '--non_interactive',
            action='store_true',
            dest='non_interactive',
            help='Do not prompt for missing parameters'
            )

        self.argparser.add_argument(
            'options',
            nargs=argparse.REMAINDER,
            help="Specific command options"
        )

        self.argparser.add_argument(
            '--force_save',
            action='store_true',
            dest='force_save',
            help='Force saving an output file even if no modification'
        )


    def parse_args(self):
        args = self.argparser.parse_args()
        return args

    @classmethod
    def printArgs(cls, args):
        print ("Arguments")
        print ("=========")
        print ("Command:", args.command)
        if args.debug:
            print (' DEBUG mode on')
        print()




"""
  Module to manage mutations
"""

import re
import sys

import biobb_model.structure_manager.model_utils as mu

ADD = 'Add'
DEL = 'Del'
MOV = 'Mov'

class MutationManager():

    def __init__(self, id_list, debug=False):
        self.mutation_list = []

        if 'file:' in id_list:
            #Load from file
            id_list = id_list.replace ('file:', '')
            print ('Reading mutation list from file ' + id_list)
            self.id_list=[]
            for line in open(id_list, 'r'):
                self.id_list.append(line.replace('\n', '').replace('\r', ''))

        else:
            self.id_list = id_list.replace(' ', '').split(',')
        #convert to list ids to Mutation instances and make the list
        self.mutation_list = list(map (Mutation, self.id_list))

    # Check mutation_list and unroll chains/models
    def prepare_mutations (self, st, debug=False, stop_on_error=True):
        for mut in self.mutation_list:
            mut.prepare(st, debug, stop_on_error)

    # Perform the modifications according to rules
    def apply_mutations (self, st, mutation_map, residue_lib, remove_H='mut', debug=False):
        mutated_res=[]
        for mut in self.mutation_list:
            mutated_res += mut.apply(st, mutation_map, residue_lib, remove_H, debug)
        return mutated_res

    def __str__(self):
        return ','.join(self.id_list)

#===============================================================================

class Mutation():

    def __init__(self, mut_id):
        if ':' not in mut_id:
            mut_id = '*:' + mut_id

        self.id = mut_id.upper()

        [self.chain, mut] = mut_id.split(':')
        mut_comps = re.match('([A-z]*)([0-9]*)([A-z]*)', mut)
        self.old_id = mu.protein_residue_check(mut_comps.group(1))
        self.new_id = mu.protein_residue_check(mut_comps.group(3))
        self.res_num = mut_comps.group(2)

        self.id = self.chain + ":" + self.old_id + self.res_num + self.new_id

    def prepare (self, st, debug=False, stop_on_error=True): # Check which mutations are possible
        if debug:
            print ('#DEBUG: Checking ' + self.id)

        self.mutations = []
        ok = 0
        for model in st.get_models():
            if self.chain == '*':
                chains = []
                for ch in model.get_list():
                    chains.append(ch.get_id())
            else:
                chains = [self.chain]

            for ch in chains:
                if int(self.res_num) in model[ch]:
                    r = model[ch][int(self.res_num)]
                    if r.get_resname() == self.old_id:
                        self.mutations.append(
                            {
                                'model':model.get_id(),
                                'chain':ch,
                                'residue':r.get_id(),
                                'new_id':self.new_id
                            })
                        ok += 1
                    else:
                        print ('#WARNING: Unknown residue ' + ch + ':' + self.old_id + self.res_num)
                else:
                    print ("#WARNING: Unknown residue " + ch + ':' + self.old_id + self.res_num)

        if ok == 0 and stop_on_error:
            print ("#ERROR: no mutations available for " + self.id)
            sys.exit(1)

    def apply(self, st, mut_map, res_lib, remove_H, debug=False):
        if debug:
            print (self.mutations)
            print ("Mutation Rules")
            print (mut_map[self.old_id][self.new_id])
        mutated_res=[]
        for m in self.mutations:
            r = st[m['model']][m['chain']][m['residue']]
            print ("Replacing " + mu.residue_id(r) + " into " + self.new_id)
# Renaming ats
            for rule in mut_map[r.get_resname()][self.new_id][MOV]:
                [old_at, new_at] = rule.split("-")
                print ("  Renaming " + old_at + " to " + new_at)
                mu.rename_atom(r, old_at, new_at)

# Deleting atoms
            for at_id in mut_map[r.get_resname()][self.new_id][DEL]:
                print ("  Deleting " + at_id)
                mu.delete_atom(r, at_id)
# Deleting H
            if remove_H == 'mut':
                mu.remove_H_from_r(r, verbose= True)

# Adding atoms (new_id required as r.resname is still the original)
            for at_id in mut_map[r.get_resname()][self.new_id][ADD]:
                print ("  Adding new atom " + at_id)
                mu.build_atom(r, at_id, res_lib, self.new_id)

#Renaming residue
            r.resname = self.new_id
            mutated_res.append(r)
        print("")
        return mutated_res



    def __str__(self):
      return self.id



#!/usr/bin/env python3

"""Module containing the FixSideChain class and the command line interface."""
import argparse
import sys
import os
from biobb_common.configuration import settings
from biobb_common.tools import file_utils as fu
import biobb_model.structure_checking.structure_checking as sc
from biobb_model.structure_checking.structure_checking import StructureChecking
from biobb_model.structure_checking.default_settings import DefaultSettings

class FixSideChain():
    """Class to model the missing atoms in aminoacid side chains of a PDB.

    Args:
        input_pdb_path (str): Input PDB file path.
        output_pdb_path (str): Output PDB file path.
        properties (dic): (None) Empty dictionary by default.
    """
    def __init__(self, input_pdb_path, output_pdb_path, properties=None, **kwargs):
        properties = properties or {}

        # Input/Output files
        self.input_pdb_path = input_pdb_path
        self.output_pdb_path = output_pdb_path

        # Properties specific for BB
        # Right now this step has not specific properties

        # Common in all BB
        self.can_write_console_log = properties.get('can_write_console_log', True)
        self.global_log = properties.get('global_log', None)
        self.prefix = properties.get('prefix', None)
        self.step = properties.get('step', None)
        self.path = properties.get('path', '')

    def launch(self):
        """Model the missing atoms in side chains."""
        options_dict = {'input_structure_path': self.input_pdb_path,
                        'options': ['--fix', 'All'],
                        'output_structure_path': self.output_pdb_path,
                        'force_save': True,
                        'res_lib_path': None,
                        'debug': False,
                        'command': 'fixside',
                        'json_output_path': None,
                        'quiet': False,
                        'data_library_path': None,
                        'non_interactive': False,
                        'check_only': False,
                        'data_dir': None}

        sets = DefaultSettings(os.path.dirname(sc.__file__))

        out_log_file_path = fu.create_name(path=self.path, prefix=self.prefix, step=self.step, name='log.out')
        fu.create_dir(os.path.dirname(os.path.abspath(out_log_file_path)))

        with open(out_log_file_path, 'w') as out_log:
            old_stdout = sys.stdout
            sys.stdout = out_log
            StructureChecking(sets, options_dict).launch()
            sys.stdout = old_stdout
        if self.can_write_console_log:
            with open(out_log_file_path, 'r') as out_log:
                print(out_log.read())


def main():
    parser = argparse.ArgumentParser(description="Model the missing atoms in aminoacid side chains of a PDB.")
    parser.add_argument('--config', required=False)
    parser.add_argument('--system', required=False)
    parser.add_argument('--step', required=False)

    # Specific args of each building block
    parser.add_argument('--input_pdb_path', required=True)
    parser.add_argument('--output_pdb_path', required=True)
    ####

    args = parser.parse_args()
    config = args.config if args.config else None
    properties = settings.ConfReader(config=config, system=args.system).get_prop_dic()
    if args.step:
        properties = properties[args.step]

    #Specific call of each building block
    FixSideChain(input_pdb_path=args.input_pdb_path, output_pdb_path=args.output_pdb_path, properties=properties).launch()
    ####

if __name__ == '__main__':
    main()




import json

class JSONWriter:
    def __init__(self):
        self.data = {}

    def set(self,k,v):
        self.data[k]=v

    def appendData(self, ar, v):
        if ar not in self.data:
            self.data[ar]=[]
        self.data[ar].append(v)

    def __str__(self):
        return json.JSONEncoder(sort_keys=True, indent=1).encode(self.data)

    def save(self,file):
        jsout = open(file,"w+")
        jsout.write(self.__str__())
        jsout.close()





"""
  module to load data library json file
"""

import json
import sys

class DataLibManager():

    def __init__(self, file_path):
        try:
            fh = open (file_path, "r")
            json_map = json.load(fh)
            self.residue_codes = json_map['data_library']['residue_codes']
            self.atom_data = json_map['data_library']['atom_data']
            self.residue_data = json_map['data_library']['residue_data']
            self.distances = json_map['data_library']['distances']

        except IOError:
            print ("ERROR: unable to open data library "+ file_path, file=sys.stderr)
            sys.exit(2)

    def get_valid_codes (self, mol_type):
        if mol_type == 'na':
            codes = self.residue_codes['dna']+ self.residue_codes['rna']
        else:
            codes = self.residue_codes[mol_type]
        return codes

    def get_all_atom_lists(self):
        ats_lists = {}
        for aa in self.residue_codes['protein']:
            ats_lists[aa] = {'backbone': self.residue_data['*']['bck_atoms'], 'side': self.residue_data[aa]['side_atoms']}
        return ats_lists

    def get_metal_atoms(self):
        return self.atom_data['metal_atoms']

    def get_atom_feature_list(self,feature):
        f_list = {}
        for rcode in self.residue_data:
            if feature in self.residue_data[rcode]:
                f_list[rcode]=self.residue_data[rcode][feature]
        if not '*' in f_list:
            f_list['*'] = []
        return f_list

    def get_atom_lists(self, contact_types):
        atom_lists={}
        for cls in contact_types:
            if cls != 'severe':
                atom_lists[cls]= self.get_atom_feature_list(cls + '_atoms')
        return atom_lists

    def get_mutation_rules(self,aa_in,aa_out,rule_group):
        return self.residue_data[aa_in]['mutation_rules'][aa_out][rule_group]

    def get_distances(self, distid):
        return self.distances[distid]

    def get_amide_data(self):
        alist=[]
        rlist={}
        for rcode in self.residue_data:
            if 'amide_atoms' in self.residue_data[rcode]:
                alist += self.residue_data[rcode]['amide_atoms']
                rlist[rcode] = self.residue_data[rcode]['amide_atoms']
        return [rlist,alist]

    def get_chiral_data(self):
        rlist={}
        for rcode in self.residue_data:
            if 'chiral_atoms' in self.residue_data[rcode]:
                rlist[rcode]=self.residue_data[rcode]['chiral_atoms']
        return rlist

    def get_mutation_map(self):
        mut_rules={}
        for rcode in self.residue_data:
            if 'mutation_rules' in self.residue_data[rcode]:
                mut_rules[rcode]=self.residue_data[rcode]['mutation_rules']
        return(mut_rules)



"""
  Utility functions to manipulate structures. based on Bio.PDB data model
"""

from Bio.PDB.Atom import Atom
from Bio.PDB.NeighborSearch import NeighborSearch
import math
import numpy as np
from numpy import arccos
from numpy import clip
from numpy import cos
from numpy import dot
from numpy import pi
from numpy import sin
from numpy.linalg import norm
import re
import sys

UNKNOWN = 0

#chain types
PROTEIN = 1
DNA = 2
RNA = 3
NA = 4
SEQ_THRESHOLD = 0.8
chain_type_labels = {PROTEIN:'Protein', DNA:'DNA', RNA:'RNA', UNKNOWN:'Unknown'}

#Model Types
ENSM = 1
BUNIT = 2
MODELS_MAXRMS = 15.0    # Threshold value to detect NMR models (angs)
model_type_labels = {ENSM:'Ensembl', BUNIT:'BioUnit', UNKNOWN:'Unknown'}


# TODO: consider replace by Bio.PDB equivalent
one_letter_residue_code = {
    'ALA':'A', 'CYS':'C', 'ASP':'D', 'GLU':'E', 'PHE':'F', 'GLY':'G',
    'HIS':'H', 'HID':'H', 'HIE':'H', 'ILE':'I', 'LYS':'K', 'LEU':'L',
    'MET':'M', 'ASN':'N', 'PRO':'P', 'GLN':'Q', 'ARG':'R', 'SER':'S',
    'THR':'T', 'VAL':'V', 'TRP':'W', 'TYR':'Y'
}

three_letter_residue_code = {
    'A':'ALA', 'C': 'CYS', 'D':'ASP', 'E':'GLU', 'F':'PHE', 'G':'GLY',
    'H':'HIS', 'I':'ILE', 'K':'LYS', 'L':'LEU', 'M':'MET', 'N':'ASN',
    'P':'PRO', 'Q':'GLN', 'R':'ARG', 'S':'SER',
    'T':'THR', 'V':'VAL', 'W':'TRP', 'Y':'TYR'
}

dna_residue_code = ['DA', 'DC', 'DG', 'DT']
rna_residue_code = ['A', 'C', 'G', 'U']
na_residue_code = dna_residue_code + rna_residue_code


# Residue ids

def residue_id(r, models='auto'):
    """
    Friendly replacement for residue ids like ASN A324/0
    """
    return '{:>3} {}'.format(r.get_resname(), residue_num(r, models))

def residue_num (r, models='auto'):
    """
    Shortcut for getting residue num including chain id, includes model number if any
    """
    if models == 'auto':
        models = len(r.get_parent().get_parent().get_parent()) > 1

    if has_ins_code(r):
        rn = str(r.get_parent().id) + str(r.id[1]) + r.id[2]
    else:
        rn = str(r.get_parent().id) + str(r.id[1])
    if models:
        rn += "/" + str(r.get_parent().get_parent().id + 1)
    return rn

def atom_id(at, models='auto'):
    """
    Friendly replacement for atom ids like ASN A324/0.CA
    """
    return '{}.{}'.format(residue_id(at.get_parent(), models), at.id)

# Id Checks
def protein_residue_check(r):
    """
    Checks whether is a valid protein residue id, either one or three letter code
    return upper case three-letter code
    """
    r = r.upper()
    rid = ''
    if r in three_letter_residue_code.keys():
        rid = three_letter_residue_code[r]
    elif r in one_letter_residue_code.keys():
        rid = r
    else:
        print ('#ERROR: unknown residue id ' + r)
        sys.exit(1)

    return rid

def same_residue (at1, at2):
    """
    Checks whether atoms belong to the same residue
    """
    return at1.get_parent() == at2.get_parent()

def same_model(r1, r2):
    """
    Checks whether residues belong to the same model
    """
    return r1.get_parent().get_parent() == r2.get_parent().get_parent()

def same_chain(r1, r2):
    """
    Checks whether residues belong to the same chain
    """
    return (r1.get_parent() == r2.get_parent()) and same_model(r1, r2)

def seq_consecutive(r1, r2):
    """
    Checks whether residues belong to the same chain and are consecutive in sequences,
    taken from residue number
    """
    resnum1 = r1.id[1]
    resnum2 = r2.id[1]
    return same_chain(r1, r2) and abs(resnum1-resnum2) == 1

def is_wat(r):
    """
    Shortcut to check for water residues
    """
    return r.id[0] == 'W'

def is_hetatm(r):
    """
    Shortcut to check for HETATM residues
    """
    return re.match('H_', r.id[0]) or re.match('W', r.id[0])

def is_at_in_list(at, at_list):
    rname = at.get_parent().get_resname().replace(' ', '')
    if not rname in at_list:
        return at.id in at_list['*']
    else:
        return at.id in at_list[rname] or at.id in at_list['*']

def has_ins_code(r):
    return r.id[2] != ' '

def guess_chain_type(ch, thres=SEQ_THRESHOLD):
    """
    Guesses chain type (protein, dna, or rna) from residue composition
    Allow for non-std residues.
    """
    #TODO improve guessing for hybrid chains
    prot = 0.
    dna = 0.
    rna = 0.
    total = 0
    for r in ch.get_residues():
        if is_wat(r):
            continue
        rname = r.get_resname().replace(' ', '')
        if rname in three_letter_residue_code.values():
            prot += 1
        elif rname in dna_residue_code:
            dna += 1
        elif rname in rna_residue_code:
            rna += 1
        total += 1
    prot = prot / total
    dna = dna / total
    rna = rna / total
    other = 1 - prot - dna - rna
    if prot > thres or prot > dna + rna:
        return PROTEIN
    elif dna > thres or dna > prot + rna:
        return DNA
    elif rna > thres or rna > prot + dna:
        return RNA
    else:
        return [prot, dna, rna, other]

def check_chiral_residue(r, chiral_data):
    """
    Checks proper chirality of side chain atoms as defined in
    chiral_data
    """
    ok = True
    if r.get_resname() in chiral_data:
        atids = chiral_data[r.get_resname()]
        ok = check_chiral(r, 'CB', 'CA', atids[0], atids[1], -1.)
    return ok

def check_chiral_ca(r):
    """
    Checks proper (L) quirality of CB atom with respect to backbone
    """
    return r.get_resname() == 'GLY' or check_chiral(r, 'N', 'CA', 'C', 'CB')

def check_chiral(r, at1, at2, at3, at4, sign=1.):
    """
    Checks proper chirality.
    at1-at3 define reference plane.
    Position of at4 with respect to the plane is checked.
    Sign (+1,-1) allows to check for a specific enantiomer.
    """
    # check all atoms are present
    at_ok = True
    chi_ok = True
    for at in [at1, at2, at3, at4]:
        at_ok = at_ok and at in r
        if not at_ok:
            print ('Warning: atom {} not found in {}'.format(at, residue_id(r)))
    if at_ok:
        v1 = r[at1].coord-r[at2].coord
        v2 = r[at3].coord-r[at2].coord
        vp = np.cross(v1, v2)
        v3 = r[at4].coord-r[at2].coord
        chi_ok = sign * (_calc_v_angle(vp, v3) - 90.) < 0.
    return chi_ok

def check_all_at_in_r(r, at_list):
    miss_at = {}
    for group in ['backbone', 'side']:
        miss_at[group] = []
        for at_id in at_list[group]:
            if not at_id in r:
                miss_at[group].append(at_id)
    if len(miss_at['backbone'] + miss_at['side']) > 0:
        return miss_at
    else:
        return {}

def get_altloc_residues(st):
    """
    Gets list of residue  with atoms with alternative location labels

    """
    res_list = {}
    for at in st.get_atoms():
        r = at.get_parent()
        if at.get_altloc() != ' ':
            if r not in res_list:
                res_list[r] = []
            res_list[r].append(at)
    return res_list

def get_metal_atoms(st, metal_ats):
    """
    Gets list of metal atoms

    """
    met_list = []
    for at in st.get_atoms():
        if not re.match('H_', at.get_parent().id[0]):
            continue
        if at.id in metal_ats:
            met_list.append(at)
    return met_list

def get_ligands(st, incl_water=False):
    """
    Gets lists of ligands, water molecules can be excluded
    """
    lig_list = []
    for r in st.get_residues():
        if re.match('H_', r.id[0]) or (incl_water and re.match('W', r.id[0])):
            lig_list.append(r)
    return lig_list

def get_residues_with_H(st):
    """
    Get residues containint Hydrogen atoms

    """
    resh_list = []
    for r in st.get_residues():
        has_h = 0
        for a in r.get_atoms():
            if a.element == 'H':
                has_h += 1
        if has_h:
            resh_list.append({'r':r, 'n_h':has_h})
    return resh_list

def check_r_list_clashes(r_list, rr_list, CLASH_DIST, atom_lists):
    clash_list = {'severe':{}}
    for cls in atom_lists:
        clash_list[cls] = {}
    for r_pair in rr_list:
        [r1, r2, d] = r_pair
        if (r1 in r_list or r2 in r_list) and not is_wat(r1) and not is_wat(r2):
            c_list = check_rr_clashes(r1, r2, CLASH_DIST, atom_lists)
            rkey = residue_id(r1) + '-' + residue_id(r2)
            for cls in c_list:
                if len(c_list[cls]):
                    clash_list[cls][rkey] = c_list[cls]
    return clash_list

def check_rr_clashes(r1, r2, CLASH_DIST, atom_lists):

    clash_list = {}
    min_dist = {}
    for cls in atom_lists:
        clash_list[cls] = []
        min_dist[cls] = 999.

    if r1 != r2 and not seq_consecutive(r1, r2) and same_model(r1, r2):
        for at_pair in get_all_rr_distances(r1, r2):
            [at1, at2, dist] = at_pair
            if 'severe' in atom_lists and dist < CLASH_DIST['severe']:
                if dist < min_dist:
                    clash_list['severe'] = at_pair
                    min_dist['severe'] = dist
            else:
                for cls in atom_lists:
                    if cls == 'apolar':
                        #Only one of the atoms should be apolar
                        if not is_at_in_list(at1, atom_lists[cls]) and not is_at_in_list(at2, atom_lists[cls]):
                            continue
                        #Remove n->n+2 backbone clashes. TODO Improve
                        if abs(at1.get_parent().index - at2.get_parent().index) <= 2:
                            continue
                        #Remove Ca2+ looking like backbone CA's
                        if is_hetatm(at1.get_parent()) and at1.id == 'CA' or \
                            is_hetatm(at2.get_parent()) and at2.id == 'CA':
                                continue
                    else:
                        # Both atoms should be of the same kind
                        if not is_at_in_list(at1, atom_lists[cls]) or not is_at_in_list(at2, atom_lists[cls]):
                            continue
                    if dist < CLASH_DIST[cls]:

                        if dist < min_dist[cls]:
                            clash_list[cls] = at_pair
                            min_dist[cls] = dist
    return clash_list

def get_backbone_links(st, backbone_atoms, COVLNK):
    cov_links = []
    for m in st:
        bckats = []
        for at in st[m.id].get_atoms():
            if at.id in backbone_atoms:
                bckats.append(at)

        if bckats:
            nbsearch = NeighborSearch(bckats)

            for at1, at2 in nbsearch.search_all(COVLNK):
                if not same_residue(at1, at2):
                    cov_links.append(sorted([at1, at2], key=lambda x: x.serial_number))
        else:
            print ("Warning: No backbone atoms defined")

    return cov_links

# Residue manipulation =======================================================
def remove_H_from_r (r, verbose=False):
    """
    Removes Hydrogen atoms from given residue
    """
    H_list = []
    for at in r.get_atoms():
        if at.element == 'H':
            H_list.append(at.id)
    for at_id in H_list:
        if verbose:
            print ("  Deleting atom " + at_id)
        r.detach_child(at_id)

def remove_residue(r):
    """
    Removes residue completely
    """
    r.get_parent().detach_child(r.id)

def swap_atoms(at1, at2):
    """
    Swaps names for two given atoms. Useful to fix labelling issues
    """
    at1_id = at1.id
    at1_full_id = at1.full_id
    at1_element = at1.element
    at1_name = at1.name
    at1_fullname = at1.fullname

    at1.id = at2.id
    at1.full_id = at2.full_id
    at1.element = at2.element
    at1.name = at2.name
    at1.fullname = at2.fullname

    at2.id = at1_id
    at2.full_id = at1_full_id
    at2.element = at1_element
    at2.name = at1_name
    at2.fullname = at1_fullname

def invert_side_atoms(r, res_data): #TODO check merging with swat_atom_names
    """
     Swaps atoms according to res_data. Useful for fixing amides and chirality issues
    """
    #TODO reconstruct CD1 atom in Ile
    res_type = r.get_resname()
    if not res_type in res_data:
        sys.stderr.write('Error: {} is not a valid residue'.format(res_type))
    swap_atoms(r[res_data[res_type][0]], r[res_data[res_type][1]])

def invert_chiral_ca(r):
    """
     Inverts CA Chirality.
    """
    #TODO
    pass
    return None

# Atom management ==============================================================
def rename_atom(r, old_at, new_at):
    at = r[old_at]
    r.detach_child(at.id)
    at.id = new_at
    at.full_id = new_at
    at.element = new_at[0:1]
    at.fullname = ' ' + new_at
    r.add(at)

def delete_atom(r, at_id):
    r.detach_child(at_id)

def build_atom(r, at_id, res_lib, new_res_id):
    if at_id == 'CB':
        coords = buildCoordsCB(r)
    else:
        coords = buildCoordsOther(r, res_lib, new_res_id, at_id)

    at = Atom(
              at_id,
              coords,
              99.0,
              1.0,
              ' ',
              ' ' + at_id + ' ',
              0,
              at_id[0:1]
              )
    r.add(at)

def buildCoordsOther(r, res_lib, new_res, at_id):
    """
     Calculates cartesian coordinates for a new atom from internal coordinates definition.

    """
    resid_def = res_lib.residues[new_res]
    i = 1
    while resid_def.ats[i].id != at_id and i < len(resid_def.ats):
        i = i + 1
    if resid_def.ats[i].id == at_id:
        return buildCoords(
                           r[resid_def.ats[resid_def.ats[i].link[0]].id].get_coord(),
                           r[resid_def.ats[resid_def.ats[i].link[1]].id].get_coord(),
                           r[resid_def.ats[resid_def.ats[i].link[2]].id].get_coord(),
                           resid_def.ats[i].geom
                           )
    else:
        print ("#ERROR: Unknown target atom")
        sys.exit(1)

def buildCoordsCB(r): # Get CB from Backbone
    """
     Calculates cartesian coordinates for a new CB atom from backbone.

    """
    return buildCoords(
                       r['CA'].get_coord(),
                       r['N'].get_coord(),
                       r['C'].get_coord(),
                       [1.5, 115.5, -123.]
                       )

def buildCoords(avec, bvec, cvec, geom):
    """
     Calculates cartesian coordinates for a new atom from internal coordinates.

    """
    dst = geom[0]
    ang = geom[1] * pi / 180.
    tor = geom[2] * pi / 180.0

    v1 = avec-bvec
    v2 = avec-cvec

    n = np.cross(v1, v2)
    nn = np.cross(v1, n)

    n /= norm(n)
    nn /= norm(nn)

    n *= -sin(tor)
    nn *= cos(tor)

    v3 = n + nn
    v3 /= norm(v3)
    v3 *= dst * sin(ang)

    v1 /= norm(v1)
    v1 *= dst * cos(ang)

    return avec + v3 - v1

# Metrics =============================================================
def calc_at_dist(at1, at2):
    """
    Calculates distance between two atoms
    """
    return np.sqrt(calc_at_sq_dist(at1, at2)) #TODO look for a faster alternative

def calc_at_sq_dist(at1, at2):
    """
    Calculates distance between two atoms
    """
    v = at1.coord-at2.coord
    return dot(v, v)

def calc_bond_angle(at1, at2, at3):
    """
    Calculates angle among three atoms at1-at2-at3
    """
    v1 = at1.coord - at2.coord
    v2 = at3.coord - at2.coord
    return _calc_v_angle(v1, v2)

def calc_bond_dihedral(at1, at2, at3, at4):
    """
    Calculates dihedral angles at1-at2-at3-at4
    """
    ab = at1.coord-at2.coord
    cb = at3.coord-at2.coord
    db = at4.coord-at3.coord
    u = np.cross(ab, cb)
    v = np.cross(db, cb)
    w = np.cross(u, v)
    angle_uv = _calc_v_angle(u, v)
    angle_cbw = _calc_v_angle(cb, w)
    try:
        if angle_cbw > 0.001:
            angle_uv = -angle_uv
    except ZeroDivisionError:
        pass
    return angle_uv

def get_all_at2at_distances(st, at_ids='all', d_cutoff=0., check_models=True):
    """
    Gets a list of all at-at distances below a cutoff, at ids can be limited
    """
    if not isinstance(at_ids, list):
        at_ids = at_ids.split(',')

    dist_mat = []
    at_list = []
    d_cut2 = d_cutoff ** 2
    for at in st.get_atoms():
        if at.id in at_ids or at_ids == ['all']:
            at_list.append(at)
    for i in range(0, len(at_list)-1):
        for j in range(i + 1, len(at_list)):
            if not check_models or same_model(at_list[i].get_parent(), at_list[j].get_parent()):
                d = calc_at_sq_dist(at_list[i], at_list[j])
                if d_cutoff > 0. and d < d_cut2:
                    dist_mat.append ([at_list[i], at_list[j], d])
    return dist_mat


def get_all_r2r_distances(st, r_ids='all', d_cutoff=0.):
    # Uses distances between the first atom of each residue as r-r distance
    if not isinstance(r_ids, list):
        r_ids = r_ids.split(',')
    dist_mat = []
    r_list = []
    d_cut2 = d_cutoff ** 2
    for r in st.get_residues():
        if r.resname in r_ids or r_ids == ['all']:
            r_list.append(r)
    for i in range(0, len(r_list)-1):
        ati = r_list[i].child_list[0]
        for j in range(i + 1, len(r_list)):
            atj = r_list[j].child_list[0]
            d = calc_at_sq_dist(ati, atj)
            if d_cutoff > 0. and d < d_cut2:
                dist_mat.append ([r_list[i], r_list[j], d])
    return dist_mat

def calc_RMSd_ats (ats1, ats2):
    if len(ats1) != len(ats2):
        print ("Warning: atom lists of different length when calculating RMSd", file=sys.stderr)
    rmsd = 0
    i = 0
    while i < len(ats1) and i < len(ats2):
        d2 = calc_at_sq_dist(ats1[i], ats2[i])
        rmsd = rmsd + d2
        i = i + 1

    return (math.sqrt(rmsd / i))

def calc_RMSd_all_ats (st1, st2):
    ats1 = []
    ats2 = []

    for at in st1.get_atoms():
        ats1.append(at)
    for at in st2.get_atoms():
        ats2.append(at)

    return calc_RMSd_ats(ats1, ats2)

def get_all_rr_distances(r1, r2, with_h=False):
    dist_mat = []
    for at1 in r1.get_atoms():
        if at1.element == 'H' and not with_h:
            continue
        for at2 in r2.get_atoms():
            if at2.element == 'H' and not with_h:
                continue
            if at1.serial_number < at2.serial_number:
                dist_mat.append ([at1, at2, calc_at_dist(at1, at2)])
    return dist_mat

def guess_models_type(st, threshold=MODELS_MAXRMS):
    if len(st) == 1:
        return 0
    rmsd = calc_RMSd_all_ats(st[0], st[1])
    if len(st) > 1:
        if rmsd < threshold:
            return {'type':ENSM, 'rmsd':rmsd}
        else:
            return {'type':BUNIT, 'rmsd':rmsd}
    else:
        return {'type':UNKNOWN, 'rmsd':rmsd}

#===============================================================================
def _calc_v_angle(v1, v2, deg=True):
    angle = arccos(clip(dot(v1, v2) / norm(v1)/norm(v2),-1.,1.))
    if deg:
        angle *= 180./pi
    return angle



#!/usr/bin/env python3

"""Module containing the Mutate class and the command line interface."""
import argparse
import sys
import os
from biobb_common.configuration   import settings
from biobb_common.tools import file_utils as fu
import biobb_model.structure_checking.structure_checking as sc
from biobb_model.structure_checking.structure_checking import StructureChecking
from biobb_model.structure_checking.default_settings import DefaultSettings

class Mutate():
    """Class to mutate one aminoacid by another in a 3d structure.

    Args:
        input_pdb_path (str): Input PDB file path.
        output_pdb_path (str): Output PDB file path.
        properties (dic):
            | - **mutation_list** (*str*): ("A:Val2Ala") Mutation list in the format "Chain:WT_AA_ThreeLeterCode Resnum MUT_AA_ThreeLeterCode" (no spaces between the elements) separated by commas. If no chain is provided as chain code all the chains in the pdb file will be mutated. ie: "A:ALA15CYS"
    """
    def __init__(self, input_pdb_path, output_pdb_path, properties=None, **kwargs):
        properties = properties or {}

        # Input/Output files
        self.input_pdb_path = input_pdb_path
        self.output_pdb_path = output_pdb_path

        # Properties specific for BB
        self.mutation_list = properties.get('mutation_list', 'A:Val2Ala')

        # Common in all BB
        self.can_write_console_log = properties.get('can_write_console_log', True)
        self.global_log = properties.get('global_log', None)
        self.prefix = properties.get('prefix', None)
        self.step = properties.get('step', None)
        self.path = properties.get('path', '')

    def launch(self):
        """Model the missing atoms in the side chains."""
        options_dict = {'force_save': False,
                        'command': 'mutateside',
                        'data_dir': None,
                        'json_output_path': None,
                        'data_library_path': None,
                        'res_lib_path': None,
                        'quiet': False,
                        'output_structure_path': self.output_pdb_path,
                        'options': ['--mut', self.mutation_list],
                        'non_interactive': False,
                        'debug': False,
                        'check_only': False,
                        'input_structure_path': self.input_pdb_path}

        sets = DefaultSettings(os.path.dirname(sc.__file__))
        out_log_file_path = fu.create_name(path=self.path, step=self.step, name='log.out')
        fu.create_dir(os.path.dirname(os.path.abspath(out_log_file_path)))

        with open(out_log_file_path, 'w') as out_log:
            old_stdout = sys.stdout
            sys.stdout = out_log
            StructureChecking(sets, options_dict).launch()
            sys.stdout = old_stdout

        if self.can_write_console_log:
            with open(out_log_file_path, 'r') as out_log:
                print(out_log.read())



def main():
    """Command line interface."""
    parser = argparse.ArgumentParser(description="Model the missing atoms in aminoacid side chains of a PDB.")
    parser.add_argument('--config', required=False)
    parser.add_argument('--system', required=False)
    parser.add_argument('--step', required=False)

    # Specific args of each building block
    parser.add_argument('--input_pdb_path', required=True)
    parser.add_argument('--output_pdb_path', required=True)

    args = parser.parse_args()
    config = args.config if args.config else None
    properties = settings.ConfReader(config=config, system=args.system).get_prop_dic()
    if args.step:
        properties = properties[args.step]
    #Specific call of each building block
    Mutate(input_pdb_path=args.input_pdb_path, output_pdb_path=args.output_pdb_path, properties=properties).launch()

if __name__ == '__main__':
    main()



"""
Class to initialize global settings like data file path.
"""
class DefaultSettings():
    def __init__(self, base_dir):
        self.base_dir_path = base_dir
#Default locations
        self.help_dir_path = self.base_dir_path + "/helpdocs"
        self.data_dir_path = self.base_dir_path + "/dat"
        self.res_library_path = self.data_dir_path + "/all_amino03.in"
        self.data_library_path = self.data_dir_path + "/data_lib.json"




#!/usr/bin/env python3

"""
 structureCheck: Command line replacement for MDWeb structure check
"""

__author__ = "gelpi"
__date__ = "$13-jul-2018 15:52:55$"

import os
import sys

from biobb_model.structure_checking.cmd_line import CmdLine
from biobb_model.structure_checking.help_manager import HelpManager
from biobb_model.structure_checking.structure_checking import StructureChecking
from biobb_model.structure_checking.default_settings import DefaultSettings

def main():

    sets = DefaultSettings(os.path.dirname(__file__))
#
    help_manager = HelpManager(sets.help_dir_path)
    cmd_line = CmdLine(defaults={})
    args = cmd_line.parse_args()

    if args.command == 'commands':
        help_manager.print_help("general", header=True, pager=True)
        sys.exit(0)

    if 'help' in args.options:
        help_manager.print_help(args.command, header=True)
        sys.exit(0)

    print(sets)
    print(vars(args))
    StructureChecking(sets,vars(args)).launch()

if __name__ == "__main__":
    main()



"""
    Main Class for Structure Checking functionality

"""

__author__ = "gelpi"
__date__ = "$26-jul-2018 14:34:51$"

import sys

from biobb_model.structure_checking.help_manager import HelpManager
from biobb_model.structure_checking.json_writer import JSONWriter
from biobb_model.structure_checking.param_input import Dialog
from biobb_model.structure_checking.param_input import ParamInput

from biobb_model.structure_manager.data_lib_manager import DataLibManager
from biobb_model.structure_manager.structure_manager import StructureManager
from biobb_model.structure_manager.mutation_manager import MutationManager
from biobb_model.structure_manager.residue_lib_manager import ResidueLib

import biobb_model.structure_manager.model_utils as mu

# Interactive dialogs to complete command_line missing paramters
dialogs = Dialog()

#dialogs.add_option(command, prompt, destination, help_text, type(str))
dialogs.add_option('command_list', '--list', 'op_list', 'Command List File')
dialogs.add_option('models', '--select_model', 'select_model', 'Select model to keep', int)
dialogs.add_option('chains', '--select_chains', 'select_chains', 'Chains (All | Chain list comma separated)')
dialogs.add_option('altloc', '--select_altloc', 'select_altloc', 'Select altloc occupancy|alt_id')
dialogs.add_option('metals', '--remove', 'remove_metals', 'Remove Metal ions')
dialogs.add_option('remwat', '--remove', 'remove_wat', 'Remove Water molecules')
dialogs.add_option('ligands', '--remove', 'remove_ligands', 'Remove Ligand residues')
dialogs.add_option('remh', '--remove', 'remove_h', 'Remove Hydrogen atoms')
dialogs.add_option('amide', '--fix', 'amide_fix', 'Fix Residues (All | None | List)')
dialogs.add_option('chiral', '--fix', 'chiral_fix', 'Fix Residues (All | None | List)')
dialogs.add_option('chiral_bck', '--fix', 'chiral_fix', 'Fix Residues (All | None | List)')
dialogs.add_option('clashes', '--no_wat', 'discard_wat', 'Discard water molecules')
dialogs.add_option('fixside', '--fix', 'fix_side', 'Add missing atoms to side chains (All | None | List)')
dialogs.add_option('mutateside', '--mut', 'mut_list', 'Mutate side chains (Mutation List as [*:]arg234Thr)')

AVAILABLE_METHODS=['models','chains','inscodes','remh','remwat', 'metals','ligands','amide','chiral','chiral_bck','fixside','backbone','cistransbck','clashes']

# Main class
class StructureChecking():
    def __init__(self, sets, args):
        self.args = args
        self.sets = sets
        self.summary = {}
        self.tmp_data = {}
        self.rr_dist = []
        self.data_library = DataLibManager(sets.data_library_path)
        if not 'Notebook' in self.args:
            self.args['Notebook'] = False
        if self.args['Notebook']:
            self.args['non_interactive'] = True
            self.args['check_only'] = False

        self.CLASH_DISTS = self.data_library.get_distances('CLASH_DIST')

    def launch(self):
        help_manager= HelpManager(self.sets.help_dir_path)
        help_manager.print_help('header')

        if '-h' in self.args['options'] or '--help' in self.args['options']:
            help_manager.print_help(self.args['command'])
            sys.exit()

        if self.args['command'] == 'command_list':
            self.command_list(self.args['options'])
        elif self.args['command'] == 'checkall':
            self.checkall(self.args['options'])
        else:
            self.run_method(self.args['command'], self.args['options'])

        if not self.args['check_only']:
            if self.stm.modified or self.args['force_save']:
                if not self.stm.modified:
                    print ('Structure not modified, saving due to --force_save option')
                self._save_structure()
                self.stm.calc_stats()
                print ('Structure saved on {}'.format(self.args['output_structure_path']))
                self.stm.print_stats('Final')
                self.summary['final_stats'] = self.stm.get_stats()
            elif not self.stm.modified:
                print ('Structure not modified, not saving. Override with --force_save')

        if self.args['json_output_path'] is not None:
            json_writer = JSONWriter()
            for k in self.summary:
                json_writer.set(k, self.summary[k])
            json_writer.save(self.args['json_output_path'])
            print ('Summary data saved on {}'.format(self.args['json_output_path']))

    def run_method(self, command, opts):
        try:
            f_check = getattr(self, command + '_check')
        except AttributeError:
            print ('Error: {} command unknown or not implemented'.format(command))
            sys.exit(1)

        if not command in self.summary:
            self.summary[command] = {}

        msg = 'Running {}.'.format(command)
        if opts:
            self.summary[command]['opts'] = opts
            msg += ' Options: {} '.format(' '.join(opts))

        if not self.args['quiet']:
            print (msg)

        self._load_structure()

    #Running checking method
        self.to_fix = f_check()

        if self.args['check_only'] or opts is None or opts == '':
            if not self.args['quiet']:
                print ('Running  check_only. Nothing else to do.')
        elif self.to_fix:
            try:
                f_fix = getattr(self, command + '_fix')
            except AttributeError:
                print ('Error: {}_fix command unknown or not implemented'.format(command))
                sys.exit(1)
            if not self.args['Notebook']:
                if dialogs.exists(command):
                    opts = dialogs.get_parameter(command, opts)
                    opts = opts[dialogs.get_dialog(command)['dest']]
                else:
                    opts = ''
            f_fix(opts)

    def command_list(self, opts):
        opts = dialogs.get_parameter('command_list', opts)
        op_list = opts[dialogs.get_dialog('command_list')['dest']]

        op_list = ParamInput('Command List File', op_list, False).run()

        try:
            fh = open(op_list, "r")

        except OSError:
            print ('Error when opening file {}'.format(op_list))
            sys.exit(1)

        self._load_structure()

        i = 1
        for line in fh:
            if line == "\n" or line[0:1] == '#':
                continue
            if not self.args['quiet']:
                print ("\nStep {}: {}".format(i, line))
            data = line.split()
            command = data[0]
            opts = data[1:]
            self.run_method(command, opts)
            i += 1

        print ("Command list completed")

    def checkall(self, opts):
        self.args['check_only'] = True
        self._load_structure()
        for meth in AVAILABLE_METHODS:
            self.run_method(meth, opts)

    def print_stats(self, verbose=True):
        self._load_structure(print_stats=False)
        self.stm.print_stats()
# =============================================================================
    def models(self, opts=None):
        self.run_method('models', opts)

    def models_check(self):
        print ('{} Model(s) detected'.format(self.stm.nmodels))
        self.summary['models'] = {'nmodels': self.stm.nmodels}
        if self.stm.nmodels > 1:
            self.summary['models']['type'] = self.stm.models_type
            if self.stm.models_type['type'] == mu.ENSM:
                print ('Models superimpose, RMSd: {:8.3f} A, guessed as ensemble type (NMR / MD TRAJ)'.format(self.stm.models_type['rmsd']))
            elif self.stm.models_type['type'] == mu.BUNIT:
                print ('Models do not superimpose, RMSd: {:8.3f} A, guessed as Biounit type'.format(self.stm.models_type['rmsd']))
            else:
                print ('Models type unknown')
            return True
        else:
            if not self.args['quiet']:
                print ("Single model found")
            return False

    def models_fix(self, select_model):
        input_line = ParamInput('Select Model Num', select_model, self.args['non_interactive'])
        input_line.add_option_all ()
        input_line.add_option ('modelno', [], opt_type='int', min_val=1, max_val=self.stm.nmodels)
        [input_option, select_model] = input_line.run()

        if input_option == 'error':
            print ('Error: unknown model {}'.format(select_model), file=sys.stderr)
            self.summary['models']['error'] = 'Unknown model '+select_model
            return 1

        print ('Selecting model num. {}'.format(select_model))
        if input_option != 'all':
            self.stm.select_model(select_model)

        self.summary['models']['selected'] = select_model

# =============================================================================
    def chains (self, opts=None):
        self.run_method('chains', opts)

    def chains_check(self):
        print ('{} Chain(s) detected'.format(len(self.stm.chain_ids)))
        for ch_id in sorted(self.stm.chain_ids):
            if isinstance(self.stm.chain_ids[ch_id],list):
                print (' {}: Unknown (PROTEIN: {s[0]:4.2f} DNA: {s[1]:4.2f} RNA: {s[2]:4.2f} Other: {s[3]:4.2f})'.format(ch_id, s=self.stm.chain_ids[ch_id]))
            else:
                print (' {}: {}'.format(ch_id, mu.chain_type_labels[self.stm.chain_ids[ch_id]]))
        self.summary['chains'] = {'ids':self.stm.chain_ids}
        return len(self.stm.chains_ids) > 1

    def chains_fix(self, select_chains):
        self.summary['chains']['selected'] = {}
        input_line = ParamInput('Select chain', select_chains, self.args['non_interactive'])
        input_line.add_option_all()
        input_line.add_option('chid', sorted(self.stm.chain_ids), multiple=True, case="sensitive")
        [input_option, select_chains] = input_line.run()

        if input_option == 'error':
            print ('Error unknown selection: {}'.format(select_chains))
            self.summary['chains']['error'] = 'Unknown selection ' + select_chains
            return 1

        if input_option == 'all':
            print ('Selecting all chains')
        else:
            self.stm.select_chains(select_chains)
            print ('Selecting chain(s) {}'.format(select_chains))
            self.stm.set_chain_ids()
            self.summary['chains']['selected'] = self.stm.chain_ids

# =============================================================================
    def inscodes(self, opts=None):
        self.run_method('inscodes', opts)

    def inscodes_check(self):
        self.stm.get_ins_codes()
        if self.stm.ins_codes_list:
            print ('{} Residues with insertion codes found'.format(len(self.stm.ins_codes_list)))
            self.summary['inscodes']=[]
            for r in self.stm.ins_codes_list:
                print (mu.residue_id(r))
                self.summary['inscodes'].append(mu.residue_id(r))
        else:
            if not self.args['quiet']:
                print ("No residues with insertion codes found")

    def inscodes_fix(self,select_codes):
        pass
# =============================================================================
    def altloc(self, opts=None):
        self.run_method('altloc', opts)

    def altloc_check(self): #TODO improve output

        self.alt_loc_res = mu.get_altloc_residues(self._get_structure())

        if len(self.alt_loc_res) > 0:
            print ('Detected {} residues with alternative location labels'.format(len(self.alt_loc_res)))

            self.summary['altloc'] = {}
            self.alt_loc_rnums = []
            self.altlocs = {}
            for r in sorted(self.alt_loc_res, key=lambda x: x.index):
                rid = mu.residue_id(r)
                print(rid)
                self.alt_loc_rnums.append(mu.residue_num(r))
                self.summary['altloc'][rid] = []
                self.altlocs[r] = sorted(self.alt_loc_res[r][0].child_dict)
                for at in self.alt_loc_res[r]:
                    s = '  {:4}'.format(at.id)
                    for alt in sorted(at.child_dict):
                        s += ' {} ({:4.2f})'.format(alt, at.child_dict[alt].occupancy)
                    print (s)
                    self.summary['altloc'][rid].append({
                                                 'atom':at.id,
                                                 'loc_label':alt,
                                                 'occupancy':at.child_dict[alt].occupancy
                                                 }
                                             )
            return True
        else:
            if not self.args['quiet']:
                print ("No residues with alternative location labels detected")
            return False

    def altloc_fix(self, select_altloc):
        #Prepare the longest possible list of alternatives
        altlocs = []
        l = 0
        for r in self.altlocs:
            if len(self.altlocs[r]) > l:
                altlocs = self.altlocs[r]
                l = len(self.altlocs[r])

        input_line = ParamInput('Select alternative', select_altloc, self.args['non_interactive'])
        input_line.add_option('occup', ['occupancy'])
        input_line.add_option('altids', altlocs, case='upper')
        input_line.add_option('resnum', self.alt_loc_rnums, opt_type='pair_list', list2=altlocs, case='sensitive', multiple=True)
        [input_option, select_altloc] = input_line.run()

        if input_option == 'error':
            print ('Error: Unknown selection {} '.format(select_altloc), file=sys.stderr)
            self.summary['altloc']['error'] = "Unknown selection " + select_altloc
            return 1

        print ('Selecting location {}'.format(select_altloc))
        if input_option == 'occup' or input_option == 'altids':
            to_fix = {}
            for r in self.alt_loc_res.keys():
                to_fix[r] = {}
                to_fix[r]['ats'] = self.alt_loc_res[r]
                to_fix[r]['select'] = select_altloc
        elif input_option == 'resnum':
            to_fix = {}
            selected_rnums = {}
            for rsel in select_altloc.split(','):
                [rn, alt] = rsel.split(':')
                selected_rnums[rn] = alt
            for r in self.alt_loc_res.keys():
                rn0 = mu.residue_num(r)
                if rn0 in selected_rnums.keys():
                    to_fix[r] = {}
                    to_fix[r]['ats'] = self.alt_loc_res[r]
                    to_fix[r]['select'] = selected_rnums[rn0]

        self.summary['altloc']['selected'] = select_altloc
        for r in to_fix.keys():
            self.stm.select_altloc_residue(r, to_fix[r])

# =============================================================================
    def metals (self, opts=None):
        self.run_method('metals', opts)

    def metals_check (self):
        self.met_list = mu.get_metal_atoms(self._get_structure(), self.data_library.get_metal_atoms())

        if len(self.met_list) > 1:
            print ('{} Metal ions found'.format(len(self.met_list)))
            self.summary['metals'] = {'detected':[]}
            self.met_rids = []
            self.at_groups = {}
            for at in sorted(self.met_list, key=lambda x: x.serial_number):
                print (" {:12}", mu.atom_id(at))
                r = at.get_parent()
                self.met_rids.append(mu.residue_num(r))
                if not at.id in self.at_groups:
                    self.at_groups[at.id] = []
                self.at_groups[at.id].append(at)
                self.summary['metals']['detected'].append(mu.residue_num(r))
            return True
        else:
            if not self.args['quiet']:
                print ("No metal ions present")
            return False

    def metals_fix(self, remove_metals):
        input_sess = ParamInput("Remove", remove_metals, self.args['non_interactive'])
        input_sess.add_option_all()
        input_sess.add_option_none()
        input_sess.add_option('atids', sorted(self.at_groups), case='sensitive', multiple=True)
        input_sess.add_option('resids', self.met_rids, case='sensitive', multiple=True)
        [input_option, remove_metals] = input_sess.run()

        if input_option == "error":
            sys.stderr.write ('Error: unknown selection {}\n'.format(remove_metals))
            self.summary['metals']['error'] = 'Unknown selection ' + remove_metals
            return 1

        if input_option == 'none':
            if not self.args['quiet']:
                print ("Nothing to do")
        else:
            if input_option == 'all':
                to_remove = self.met_list
            elif input_option == 'resids':
                to_remove = []
                rid_list = remove_metals.split(',')
                for at in self.met_list:
                    r = at.get_parent()
                    if mu.residue_num(r) in rid_list:
                        to_remove.append(at)
            elif input_option == 'atids':
                to_remove = []
                for atid in remove_metals.split(','):
                    to_remove.extend(self.at_groups[atid])

            self.summary['metals']['removed'] = []

            n = 0
            for at in to_remove:
                self.summary['metals']['removed'].append(mu.residue_id(at.get_parent(), self.stm.nmodels > 1))
                self.stm.remove_residue(at.get_parent())
                n += 1

            print ('Metal Atoms removed {} ({:d})'.format(remove_metals, n))
            self.summary['metals']['n_removed'] = n

# =============================================================================
    def remwat(self, opts=None):
        self.run_method('remwat', opts)

    def remwat_check(self):
        self.lig_list = mu.get_ligands(self._get_structure(), incl_water=True)
        self.wat_list = []
        for r in self.lig_list:
            if mu.is_wat(r):
                self.wat_list.append(r)

        if len(self.wat_list) > 0:
            print ('{} Water molecules detected'.format(len(self.wat_list)))
            self.summary['remwat']['n_detected'] = len(self.wat_list)
            return True
        else:
            if not self.args['quiet']:
                print ("No water molecules found")
            return False

    def remwat_fix(self, remove_wat):
        input_line = ParamInput('Remove', remove_wat, self.args['non_interactive'])
        input_line.add_option_yes_no()
        [input_option, remove_wat] = input_line.run()

        if input_option == 'error':
            print ('Warning: unknown option {}'.format(remove_wat))
            self.summary['remwat']['error'] = 'Unknown option'
            return 1

        if input_option == 'yes':
            n = 0
            for r in self.wat_list:
                self.stm.remove_residue(r)
                n += 1
            print ('{} Water molecules removed'.format(n))
            self.summary['remwat']['n_removed'] = n
            self.lig_list = mu.get_ligands(self._get_structure(), incl_water=True)

# =============================================================================
    def ligands(self, opts=None):
        self.run_method('ligands', opts)

    def ligands_check(self):
        self.lig_list = mu.get_ligands(self._get_structure(), incl_water=False)

        if len(self.lig_list):
            print ('{} Ligands detected '.format(len(self.lig_list)))
            self.ligand_rids = set()
            self.ligand_rnums = []
            self.summary['ligands'] = {'detected': []}

            for r in sorted(self.lig_list, key=lambda x: x.index):
                if self.stm.has_models():
                    if r.get_parent().get_parent().id > 0:
                        continue
                    print (' '+mu.residue_id(r,False)+'/*')
                else:
                    print (' '+mu.residue_id(r,False))
                self.summary['ligands']['detected'].append(mu.residue_id(r, self.stm.has_models()))
                self.ligand_rids.add(r.get_resname())
                self.ligand_rnums.append(mu.residue_num(r))
            return True
        else:
            if not self.args['quiet']:
                print ("No ligands found")
            return False

# =============================================================================
    def ligands_fix(self, remove_ligands):
        input_line = ParamInput('Remove', remove_ligands, self.args['non_interactive'])
        input_line.add_option_all()
        input_line.add_option_none()
        input_line.add_option('byrids', sorted(self.ligand_rids), multiple=True)
        input_line.add_option('byresnum', self.ligand_rnums, case='sensitive', multiple=True)
        [input_option, remove_ligands] = input_line.run()

        if input_option == 'error':
            print ('Error: unknown selection {}'.format(remove_ligands))
            self.summary['ligands']['error'] = 'Unknown selection'
            return 1

        self.summary['ligands']['removed'] = {'opt':remove_ligands, 'lst':[]}

        to_remove = []

        if input_option == 'none':
            if not self.args['quiet']:
                print ("Nothing to do")
        else:
            if input_option == 'all':
                to_remove = self.lig_list
            elif input_option == 'byrids':
                rm = remove_ligands.split(',')
                for r in self.lig_list:
                    if r.get_resname() in rm:
                        to_remove.append(r)
            elif input_option == 'byresnum':
                rm = remove_ligands.split(',')
                for r in self.lig_list:
                    if mu.residue_num(r) in rm:
                        to_remove.append(r)
            n = 0
            for r in to_remove:
                self.summary['ligands']['removed']['lst'].append(mu.residue_id(r, self.stm.nmodels > 1))
                self.stm.remove_residue(r)
                n += 1

            print ('Ligands removed {} ({})'.format(remove_ligands, n))
            self.summary['ligands']['n_removed'] = n


# =============================================================================
    def remh(self, opts=None):
        self.run_method('remh', opts)

    def remh_check(self):
        self.remh_list = mu.get_residues_with_H(self._get_structure())

        if len(self.remh_list) > 0:
            print ('{} Residues containing H atoms detected'.format(len(self.remh_list)))
            self.summary['remh']['n_detected'] = len(self.remh_list)
            return True
        else:
            if not self.args['quiet']:
               print ("No residues with Hydrogen atoms found")
            return False

    def remh_fix(self, remove_h):
        input_line = ParamInput('Remove hydrogen atoms', remove_h, self.args['non_interactive'])
        input_line.add_option_yes_no()
        [input_option, remove_h] = input_line.run()

        if input_option == 'error':
            print ('Warning: unknown option {}'.format(remove_h))
            self.summary['remh']['error'] = 'Unknown option'
            return 1
        if input_option == 'yes':
            n = 0
            for r in self.remh_list:
                mu.remove_H_from_r(r['r'])
                n += 1
            print ('Hydrogen atoms removed from {} residues'.format(n))
            self.stm.modified = True
            self.summary['remh']['n_removed'] = n

# =============================================================================
    def getss (self, opts=None):
        self.run_method('getss', opts)

    def getss_check(self):
        self.SS_bonds = mu.get_all_at2at_distances(self._get_structure(), 'SG', self.data_library.get_distances('SS_DIST'))

        if len(self.SS_bonds):
            print ('{} Possible SS Bonds detected'.format(len(self.SS_bonds)))
            self.summary['getss'] = []
            for ssb in self.SS_bonds:
                print (' {:12} {:12} {:8.3f}'.format(mu.atom_id(ssb[0]), mu.atom_id(ssb[1]), ssb[2]))
                self.summary['getss'].append({'at1':mu.atom_id(ssb[0]), 'at2':mu.atom_id(ssb[1]), 'dist': float(ssb[2])})
        else:
            if not self.args['quiet']:
                print ("No SS bonds detected")

        return False

    def getss_fix(self):
        pass

# =============================================================================
    def amide(self, opts=None):
        self.run_method('amide', opts)

    def amide_check(self):

        [amide_res, amide_atoms] = self.data_library.get_amide_data()
        contact_types = ['polar_donor','polar_acceptor']
        atom_lists = self.data_library.get_atom_lists(contact_types)

        self.amide_list = []

        for r in self._get_structure().get_residues():
            if r.get_resname() in amide_res:
                self.amide_list.append(r)

        self.summary['amide']['n_amides'] = len(self.amide_list)

        if len(self.amide_list) > 0:
            if len(self.rr_dist) == 0:
                self.rr_dist = mu.get_all_r2r_distances(self._get_structure(), 'all', self.data_library.get_distances('R_R_CUTOFF'))
            self.amide_res_to_fix = []
            self.amide_cont_list = []
            self.amide_rnums = []

            c_list = mu.check_r_list_clashes(self.amide_list, self.rr_dist, self.CLASH_DISTS, atom_lists)
            for cls in c_list:
                for rkey in c_list[cls]:
                    [at1,at2,d]=c_list[cls][rkey]
                    r1 = at1.get_parent()
                    r2 = at2.get_parent()
                    add_pair=False
                    if at1.id in amide_atoms and r1 in self.amide_list:
                        self.amide_res_to_fix.append(r1)
                        self.amide_rnums.append(mu.residue_num(r1))
                        add_pair=True
                    if at2.id in amide_atoms and r2 in self.amide_list:
                        self.amide_res_to_fix.append(r2)
                        self.amide_rnums.append(mu.residue_num(r2))
                        add_pair=True
                    if add_pair:
                        self.amide_cont_list.append(c_list[cls][rkey])

            if len(self.amide_cont_list):
                print ('{} unusual contact(s) involving amide atoms found'.format(len(self.amide_cont_list)))
                self.summary['amide']['detected'] = []
                for at_pair in sorted(self.amide_cont_list, key=lambda x: x[0].serial_number):
                    print (' {:12} {:12} {:8.3f} A'.format(mu.atom_id(at_pair[0]), mu.atom_id(at_pair[1]), at_pair[2]))
                    self.summary['amide']['detected'].append({'at1':mu.atom_id(at_pair[0]), 'at2':mu.atom_id(at_pair[1]), 'dist': float(at_pair[2])})
                return True
            else:
                if not self.args['quiet']:
                    print ("No unusual contact(s) involving amide atoms found")
                return False
        else:
            if not self.args['quiet']:
                print ("No amide residues found")
            return False

# =============================================================================
    def amide_fix(self, amide_fix, recheck=True):
        amide_res = self.data_library.get_amide_data()[0]

        no_int_recheck = amide_fix is not None or self.args['non_interactive']

        fix=True
        while fix:
            input_line = ParamInput('Fix amide atoms', amide_fix, self.args['non_interactive'])
            input_line.add_option_all()
            input_line.add_option_none()
            input_line.add_option('resnum', self.amide_rnums, case='sensitive', multiple=True)
            [input_option, amide_fix] = input_line.run()

            if input_option == 'error':
                print ('Warning: unknown option {}'.format(amide_fix))
                self.summary['amide']['error'] = 'Unknown option'
                return 1

            if input_option == 'none':
                if not self.args['quiet']:
                    print ("Nothing to do")
                fix=False
            else:
                if input_option == 'all':
                    to_fix = self.amide_res_to_fix
                else:
                    to_fix = []
                    for r in self.amide_res_to_fix:
                        if mu.residue_num(r) in amide_fix.split(','):
                            to_fix.append(r)
                n = 0
                done=[]
                for r in to_fix:
                    if r not in done:
                        mu.invert_side_atoms(r, amide_res)
                        done.append(r)
                    n += 1
                print ('Amide residues fixed {} ({})'.format(amide_fix, n))
                fix = False
                if recheck:
                    if not self.args['quiet']:
                        print ("Rechecking")
                    fix=self.amide_check() #TODO reduce check to fixed residues if necessary
                    amide_fix=''
                    if no_int_recheck:
                        fix=False
                self.stm.modified = True

# =============================================================================
    def chiral(self, opts=None):
        self.run_method('chiral', opts)

    def chiral_check(self):

        chiral_res = self.data_library.get_chiral_data()

        self.chiral_list = []

        for r in self._get_structure().get_residues():
            if r.get_resname() in chiral_res:
                self.chiral_list.append(r)

        self.summary['chiral']['n_chirals'] = len(self.chiral_list)

        if len(self.chiral_list) > 0:
            self.chiral_res_to_fix = []
            self.chiral_rnums = []
            for r in self.chiral_list:
                if not mu.check_chiral_residue(r, chiral_res):
                    self.chiral_res_to_fix.append(r)
                    self.chiral_rnums.append(mu.residue_num(r))
            if len(self.chiral_res_to_fix):
                print ('{} residues with incorrect side-chain chirality found'.format(len(self.chiral_res_to_fix)))
                self.summary['chiral']['detected'] = []
                for r in self.chiral_res_to_fix:
                    print (' {:10}'.format(mu.residue_id(r)))
                    self.summary['chiral']['detected'].append(mu.residue_id(r))
                return True
            else:
                if not self.args['quiet']:
                       print ("No residues with incorrect side-chain chirality found")
                return False
        else:
            if not self.args['quiet']:
                print ("No chiral side-chains found")
            return False

    def chiral_fix(self, chiral_fix, check_clashes=True):

        chiral_res = self.data_library.get_chiral_data()

        input_line = ParamInput('Fix chiralities', chiral_fix, self.args['non_interactive'])
        input_line.add_option_all()
        input_line.add_option_none()
        input_line.add_option('resnum', self.chiral_rnums, case='sensitive', multiple=True)
        [input_option, chiral_fix] = input_line.run()

        if input_option == 'error':
            print ('Warning: unknown option {}'.format(chiral_fix))
            self.summary['chiral']['error'] = 'Unknown option'
            return 1

        if input_option == 'none':
            if not self.args['quiet']:
                print ("Nothing to do")
        else:
            if input_option == 'all':
                to_fix = self.chiral_res_to_fix
            else:
                to_fix = []
                for r in self.chiral_res_to_fix:
                    if mu.residue_num(r) in chiral_fix.split(','):
                        to_fix.append(r)
            n = 0
            for r in to_fix:
                mu.invert_side_atoms(r, chiral_res)
                if r.get_resname() == 'ILE':
                    mu.delete_atom(r,'CD1')
                    if not hasattr(self,'residue_lib'):
                        self.residue_lib = ResidueLib(self.sets.res_library_path)
                    mu.build_atom(r, 'CD1', self.residue_lib,'ILE')
                n += 1
            print ('Quiral residues fixed {} ({})'.format(chiral_fix, n))

            if check_clashes:
                if not self.args['quiet']:
                    print ("Checking for steric clashes")

                contact_types = ['severe','apolar','polar_acceptor','polar_donor', 'positive', 'negative']
                atom_lists = self.data_library.get_atom_lists(contact_types)

                if len(self.rr_dist) == 0:
                    self.rr_dist = mu.get_all_r2r_distances(self._get_structure(), 'all', self.data_library.get_distances('R_R_CUTOFF'))

                self.summary['chiral_clashes'] = self._clash_report(
                    contact_types,
                    mu.check_r_list_clashes(to_fix, self.rr_dist, self.CLASH_DISTS, atom_lists)
                )

            self.stm.modified = True

# =============================================================================
    def chiral_bck(self, opts=None):
        self.run_method('chiral_bck', opts)

    def chiral_bck_check(self):

        self.chiral_bck_list = []
        prot_chains=0
        for ch in self._get_structure().get_chains():
            if self.stm.chain_ids[ch.id] == mu.PROTEIN:
                prot_chains += 1
                for r in ch.get_residues():
                    if r.get_resname() != 'GLY' and not mu.is_hetatm(r):
                        self.chiral_bck_list.append(r)
        self.summary['chiral_bck']['n_chirals'] = len(self.chiral_bck_list)

        if len(self.chiral_bck_list) > 0:
            self.chiral_bck_res_to_fix = []
            self.chiral_bck_rnums = []
            for r in self.chiral_bck_list:
                if not mu.check_chiral_ca(r):
                    self.chiral_bck_res_to_fix.append(r)
                    self.chiral_bck_rnums.append(mu.residue_num(r))
            if len(self.chiral_bck_res_to_fix):
                print ('{} residues with incorrect backbone chirality found'.format(len(self.chiral_bck_res_to_fix)))
                self.summary['chiral_bck']['detected'] = []
                for r in self.chiral_bck_res_to_fix:
                    print (' {:10}'.format(mu.residue_id(r)))
                    self.summary['chiral_bck']['detected'].append(mu.residue_id(r))
                return False
            else:
                if not self.args['quiet']:
                    print ("No residues with incorrect backbone chirality found")
                return False
        else:
            if not prot_chains:
                print ("No protein chains detected, skipping")
            else:
                print ("Protein chains detected but no residues with chiral CA found (weird!!)")
                if not self.args['quiet']:
                    print ("No residues with chiral CA found (weird!!)")
            return False


    def chiral_bck_fix(self, chiral_fix, check_clashes=True):
        return 0 #TODO
        input_line = ParamInput('Fix CA chiralities', chiral_fix, self.args['non_interactive'])
        input_line.add_option_all()
        input_line.add_option_none()
        input_line.add_option('resnum', self.chiral_bck_rnums, case='sensitive', multiple=True)
        [input_option, chiral_fix] = input_line.run()
        if input_option == 'error':
            print ('Warning: unknown option {}'.format(chiral_fix))
            self.summary['chiral']['error'] = 'Unknown option'
            return 1

        if input_option == 'none':
            if not self.args['quiet']:
                print ("Nothing to do")
        else:
            if input_option == 'all':
                to_fix = self.chiral_bck_res_to_fix
            else:
                to_fix = []
                for r in self.chiral_bck_res_to_fix:
                    if mu.residue_num(r) in chiral_fix.split(','):
                        to_fix.append(r)
            n = 0
            for r in to_fix:
                mu.invert_chiral_ca(r)
                n += 1
            print ('Quiral residues fixed {} ({})'.format(chiral_fix, n))
            if check_clashes:
                if not self.args['quiet']:
                    print ("Checking new steric clashes")
                #TODO

            self.stm.modified = True

# =============================================================================
    def clashes(self, opts=None):
        self.run_method('clashes', opts)

    def clashes_check(self):
        contact_types = ['severe','apolar','polar_acceptor','polar_donor', 'positive', 'negative']

        self.clash_list={}
        atom_lists = self.data_library.get_atom_lists(contact_types)
        for cls in contact_types:
            self.clash_list[cls]={}

        if len(self.rr_dist) == 0:
            self.rr_dist = mu.get_all_r2r_distances(self._get_structure(), 'all', self.data_library.get_distances('R_R_CUTOFF'))

        for r_pair in self.rr_dist:
            [r1, r2, d] = r_pair
            if mu.is_wat(r1) or mu.is_wat(r2):
                continue
            c_list = mu.check_rr_clashes(r1,r2,self.CLASH_DISTS, atom_lists)
            rkey = mu.residue_id(r1)+'-'+mu.residue_id(r2)
            for cls in c_list:
                if len(c_list[cls]):
                    self.clash_list[cls][rkey] = c_list[cls]

        self.summary['clashes']['detected'] = self._clash_report(contact_types, self.clash_list)

        return False

    def clashes_fix(self, res_to_fix):
        pass
# =============================================================================
    def fixside (self, opts=None):
        self.run_method('fixside', opts)

    def fixside_check (self):
        self.miss_at_list = self.stm.get_missing_side_chain_atoms(self.data_library.get_valid_codes('protein'), self.data_library.get_all_atom_lists())

        if len(self.miss_at_list):
            self.fixside_rnums = []
            self.summary['fixside']['detected'] = {}
            print('{} Residues with missing side chain atoms found'.format(len(self.miss_at_list)))
            for r_at in self.miss_at_list:
                [r, at_list] = r_at
                print (' {:10} {}'.format(mu.residue_id(r), ','.join(at_list)))
                self.fixside_rnums.append(mu.residue_num(r))
                self.summary['fixside']['detected'][mu.residue_id(r)] = at_list
            return True
        else:
            if not self.args['quiet']:
                print ("No residues with missing side chain atoms found")
            return False

    def fixside_fix(self, fix_side, check_clashes=True):
        input_line = ParamInput ('fixside', fix_side, self.args['non_interactive'])
        input_line.add_option_all()
        input_line.add_option_none()
        input_line.add_option('resnum', self.fixside_rnums, case='sensitive', multiple=True)
        [input_option, fix_side] = input_line.run()

        if input_option == 'error':
            print ("Invalid option", fix_side)
            self.summary['fix_side']['error'] = "Unknown option"
            return 1

        self.summary['fixside']['selected'] = fix_side

        if input_option == 'none':
            if not self.args['quiet']:
                print ('Nothing to do')
        else:
            if input_option == 'all':
                to_fix = self.miss_at_list
            else:
                to_fix = []
                for r_at in self.miss_at_list:
                    if mu.residue_num(r_at[0]) in fix_side.split(','):
                        to_fix.append(r_at)
            if not self.args['quiet']:
                print ("Fixing side chains")
            n = 0
            self.summary['fixside']['fixed'] = []
            if not hasattr(self,'residue_lib'):
                self.residue_lib = ResidueLib(self.sets.res_library_path)
            fixed_res=[]
            for r_at in to_fix:
                print (mu.residue_id(r_at[0]))
                mu.remove_H_from_r(r_at[0], verbose= True)
                self.stm.fix_side_chain(r_at, self.residue_lib)
                n += 1
                self.summary['fixside']['fixed'].append(mu.residue_id(r_at[0]))
                fixed_res.append(r_at[0])

            print ('Fixed {} side chain(s)'.format(n))
            # Checking new clashes
            if check_clashes:
                print ("Checking possible clashes")
                contact_types = ['severe','apolar','polar_acceptor','polar_donor', 'positive', 'negative']
                atom_lists= self.data_library.get_atom_lists(contact_types)

                if len(self.rr_dist) == 0:
                    self.rr_dist = mu.get_all_r2r_distances(self._get_structure(), 'all', self.data_library.get_distances('R_R_CUTOFF'))

                self.summary['fixside_clashes'] = self._clash_report(
                    contact_types,
                    mu.check_r_list_clashes(fixed_res, self.rr_dist, self.CLASH_DISTS, atom_lists)
                )


# =============================================================================
    def mutateside(self, mut_list):
        self.run_method('mutateside', mut_list)

    def mutateside_check(self):
        return True

    def mutateside_fix(self, mut_list, check_clashes=True):
        input_line = ParamInput ('Mutation list', mut_list, self.args['non_interactive'])
        mut_list = input_line.run()
        self.mutations = MutationManager(mut_list)
        self.mutations.prepare_mutations(self.stm.st)
        print ('Mutations to perform')
        for m in self.mutations.mutation_list:
            print (m)
        if not hasattr(self,'mutation_rules'):
            self.mutation_rules = self.data_library.get_mutation_map()

        if not hasattr(self,'residue_lib'):
            self.residue_lib = ResidueLib(self.sets.res_library_path)

        mutated_res = self.mutations.apply_mutations (self.stm.st, self.mutation_rules, self.residue_lib)

        if check_clashes:
            if not self.args['quiet']:
                print ("Checking new clashes")

            contact_types = ['severe','apolar','polar_acceptor','polar_donor', 'positive', 'negative']
            atom_lists= self.data_library.get_atom_lists(contact_types)

            if len(self.rr_dist) == 0:
                self.rr_dist = mu.get_all_r2r_distances(self._get_structure(), 'all', self.data_library.get_distances('R_R_CUTOFF'))

            self.summary['mutateside_clashes'] = self._clash_report(
                contact_types,
                mu.check_r_list_clashes(mutated_res, self.rr_dist, self.CLASH_DISTS, atom_lists)
            )

        self.stm.modified=True
#===============================================================================
    def backbone(self, opts):
        self.run_method('backbone',opts)

    def backbone_check(self):
        backbone_atoms = self.data_library.get_all_atom_lists()['GLY']['backbone']
        # Residues with missing backbone
        self.miss_at_list = self.stm.get_missing_backbone_atoms(self.data_library.get_valid_codes('protein'), self.data_library.get_all_atom_lists())
        if len(self.miss_at_list):
            self.summary['backbone']['missing_atoms'] = {}
            print('{} Residues with missing backbone atoms found'.format(len(self.miss_at_list)))
            for r_at in self.miss_at_list:
                [r, at_list] = r_at
                print (' {:10} {}'.format(mu.residue_id(r), ','.join(at_list)))
                self.summary['backbone']['missing_atoms'][mu.residue_id(r)] = at_list
        self.stm.check_backbone_connect(backbone_atoms, self.data_library.get_distances('COVLNK'))
        #Not bound consecutive residues
        self.stm.get_bck_breaks()
        if self.stm.bck_breaks_list:
            print ("{} Backbone breaks found".format(len(self.stm.bck_breaks_list)))
            self.summary['backbone']['breaks']=[]
            for br in self.stm.bck_breaks_list:
                print (" {:10} - {:10} ".format(mu.residue_id(br[0]),mu.residue_id(br[1])))
                self.summary['backbone']['breaks'].append([mu.residue_id(br[0]),mu.residue_id(br[1])])
        if self.stm.wrong_link_list:
            print ("Unexpected backbone links found")
            self.summary['backbone']['wrong_links']=[]
            for br in self.stm.wrong_link_list:
                print (" {:10} linked to {:10}, expected {:10} ".format(mu.residue_id(br[0]),mu.residue_id(br[1]),mu.residue_id(br[2])))
                self.summary['backbone']['wrong_links'].append([mu.residue_id(br[0]),mu.residue_id(br[1]),mu.residue_id(br[2])])
        if self.stm.not_link_seq_list:
            print ("Consecutive residues too far away to be covalently linked")
            self.summary['backbone']['long_links']=[]
            for br in self.stm.not_link_seq_list:
                print (" {:10} - {:10}, bond distance {:8.3f} ".format(mu.residue_id(br[0]),mu.residue_id(br[1]),br[2]))
                self.summary['backbone']['long_links'].append([mu.residue_id(br[0]),mu.residue_id(br[1]),br[2]])
        #TODO move this section to ligands
        if self.stm.modified_residue_list:
            print ("Modified residues found")
            self.summary['backbone']['mod_residues']=[]
            for br in self.stm.modified_residue_list:
                print (" {:10} ".format(mu.residue_id(br)))
                self.summary['backbone']['mod_residues'].append(mu.residue_id(br))


    def backbone_fix(self, fix):
        #TODO
        pass
#===============================================================================
    def cistransbck(self, opts):
        self.run_method('cistransbck',opts)

    def cistransbck_check(self):
        backbone_atoms = self.data_library.get_all_atom_lists()['GLY']['backbone']
        self.stm.check_cis_backbone(backbone_atoms, self.data_library.get_distances('COVLNK'))
        if self.stm.cis_backbone_list:
            print ('{} cis peptide bonds'.format(len(self.stm.cis_backbone_list)))
            for lnk in self.stm.cis_backbone_list:
                [r1,r2,dih] = lnk
                print ('{:10} {:10} Dihedral: {:8.3f}'.format(mu.residue_id(r1), mu.residue_id(r2), dih))
        else:
            if not self.args['quiet']:
                print ("No cis peptide bonds found")

        if self.stm.lowtrans_backbone_list:
            print ('{} trans peptide bonds with unusual omega dihedrals'.format(len(self.stm.lowtrans_backbone_list)))
            for lnk in self.stm.lowtrans_backbone_list:
                [r1,r2,dih] = lnk
                print ('{:10} {:10} Dihedral: {:8.3f}'.format(mu.residue_id(r1), mu.residue_id(r2), dih))
        else:
            if not self.args['quiet']:
                print ("No trans peptide bonds with unusual omega dihedrals found")

    def cistransbck_fix(self,option):
        pass
#===============================================================================

    def _load_structure(self, verbose=True, print_stats=True):
        if not hasattr(self, 'stm'):
            if not self.args['non_interactive'] and self.args['input_structure_path'] is None:
                self.args['input_structure_path'] = input("Enter input structure path (PDB, mmcif | pdb:pdbid): ")

            self.stm = StructureManager(self.args['input_structure_path'])
            if verbose:
                print ('Structure {} loaded'.format(self.args['input_structure_path']))
                self.stm.print_headers()
                print()
            if print_stats:
                self.stm.print_stats()
                print()
            self.summary['stats'] = self.stm.get_stats()

    def _get_structure(self):
        return self.stm.get_structure()

    def _save_structure(self):
        if not self.args['non_interactive'] and self.args['output_structure_path'] is None:
            self.args['output_structure_path'] = input("Enter output structure path: ")
        self.stm.save_structure(self.args['output_structure_path'])

    def _clash_report(self, contact_types, clash_list):
        summary={}
        for cls in contact_types:
            summary[cls]=[]
            if len(clash_list[cls]):
                print ('{} Steric {} clashes detected'.format(len(clash_list[cls]), cls))
                for rkey in sorted(clash_list[cls], key=lambda x: clash_list[cls][x][0].serial_number):
                    print (' {:12} {:12} {:8.3f} A'.format(mu.atom_id(clash_list[cls][rkey][0]), mu.atom_id(clash_list[cls][rkey][1]), clash_list[cls][rkey][2]))
                    summary[cls].append(
                        {
                            'at1':mu.atom_id(clash_list[cls][rkey][0]),
                            'at2':mu.atom_id(clash_list[cls][rkey][1]),
                            'dist': float(clash_list[cls][rkey][2])
                        }
                    )
            else:
                if not self.args['quiet']:
                    print ('No {} clashes detected'.format(cls))
        return summary

#===============================================================================



"""
   Simple module for managing help texts
"""
import os
import pydoc
import sys

class HelpManager():

    def __init__(self, help_dir_path):
        self.help_dir_path = help_dir_path
        self.files = os.listdir(help_dir_path)

    def get(self, what, header=False):
        s = ''
        if header:
            if 'header.hlp' in self.files:
                with open(self.help_dir_path + "/header.hlp") as f:
                    s += f.read()
        if what + '.hlp' in self.files:
            with open(self.help_dir_path + "/" + what + ".hlp") as f:
                s += f.read()
        else:
            print ("Error: no help available for " + what)
            sys.exit()

        return s

    def print_help(self, what, header=False, pager=False):
        if pager:
            pydoc.pager(self.get(what, header))
        else:
            print (self.get(what, header))
